@@ -1,5 +1,5 @@
 /*
- * card.c: General SmartCard functions
+ * card.c: General smart card functions
  *
  * Copyright (C) 2001, 2002  Juha Yrjl <juha.yrjola@iki.fi>
  *
@@ -22,15 +22,21 @@
 #include "asn1.h"
 #include <assert.h>
 #include <stdlib.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 #include <string.h>
 
-int sc_check_sw(struct sc_card *card, int sw1, int sw2)
+int sc_check_sw(sc_card_t *card, unsigned int sw1, unsigned int sw2)
 {
-	assert(card->ops->check_sw != NULL);
+	if (card == NULL)
+		return SC_ERROR_INVALID_ARGUMENTS;
+	if (card->ops->check_sw == NULL)
+		return SC_ERROR_NOT_SUPPORTED;
 	return card->ops->check_sw(card, sw1, sw2);
 }
 
-static int sc_check_apdu(struct sc_context *ctx, const struct sc_apdu *apdu)
+static int sc_check_apdu(sc_context_t *ctx, const sc_apdu_t *apdu)
 {
 	if (apdu->le > 256) {
 		sc_error(ctx, "Value of Le too big (maximum 256 bytes)\n");
@@ -99,7 +105,7 @@ sc_masquerade_apdu(sc_card_t *card, sc_apdu_t *apdu)
 	sc_context_t	*ctx = card->reader->ctx;
 	int	masq = card->reader->driver->apdu_masquerade;
 	int	is_t0;
-	
+
 	is_t0 = (card->slot->active_protocol == SC_PROTO_T0);
 
 	if (apdu->cse == SC_APDU_CASE_4_SHORT
@@ -123,7 +129,9 @@ sc_masquerade_apdu(sc_card_t *card, sc_apdu_t *apdu)
 	return 0;
 }
 
-static int sc_transceive(struct sc_card *card, struct sc_apdu *apdu)
+/** Builds the TPDU and sends it to the reader driver
+ */
+static int sc_transceive(sc_card_t *card, sc_apdu_t *apdu)
 {
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
@@ -155,6 +163,9 @@ static int sc_transceive(struct sc_card *card, struct sc_apdu *apdu)
 	*data++ = apdu->p2;
 	switch (apdu->cse) {
 	case SC_APDU_CASE_1:
+		if (card->slot->active_protocol == SC_PROTO_T0)
+			/* T0 adds an additional 0x00 byte to the TPDU */
+			*data++ = 0x00;
 		break;
 	case SC_APDU_CASE_2_SHORT:
 		*data++ = (u8) apdu->le;
@@ -177,20 +188,34 @@ static int sc_transceive(struct sc_card *card, struct sc_apdu *apdu)
 			return SC_ERROR_INVALID_ARGUMENTS;
 		memcpy(data, apdu->data, data_bytes);
 		data += data_bytes;
-		if (apdu->le == 256)
-			*data++ = 0x00;
-		else
-			*data++ = (u8) apdu->le;
+		if (card->slot->active_protocol != SC_PROTO_T0)
+			/* unless T0 is used add Le byte */
+			*data++ = (u8) (apdu->le & 0xff);
 		break;
 	}
 	sendsize = data - sbuf;
+#if 0
 	recvsize = apdu->resplen + 2;	/* space for the SW's */
+#else
+	/* workaround for a broken cashmouse driver, should be harmless
+	 * for other drivers -- Nils */
+	recvsize = sizeof(rbuf);
+#endif
 	if (card->ctx->debug >= 5) {
 		char buf[2048];
 
-		buf[0] = 0;
+#ifndef OPENSC_DONT_LOG_SENSITIVE
 		if (!apdu->sensitive || card->ctx->debug >= 6)
+#else
+		if (!apdu->sensitive)
+#endif
 			sc_hex_dump(card->ctx, sbuf, sendsize, buf, sizeof(buf));
+		else
+			/* sensitive information: just print the command
+			 * header and no data */
+			snprintf(buf, sizeof(buf), "%02x %02x %02x %02x [sensitive data]",
+				apdu->cla, apdu->ins, apdu->p1, apdu->p2);
+
 		sc_debug(card->ctx, "Sending %d bytes (resp. %d bytes%s):\n%s",
 			sendsize, recvsize,
 			apdu->sensitive ? ", sensitive" : "", buf);
@@ -199,7 +224,7 @@ static int sc_transceive(struct sc_card *card, struct sc_apdu *apdu)
 					sendsize, rbuf, &recvsize,
 					apdu->control);
 	if (apdu->sensitive)
-		memset(sbuf, 0, sendsize);
+		sc_mem_clear(sbuf, sendsize);
 	SC_TEST_RET(card->ctx, r, "Unable to transmit");
 
 	assert(recvsize >= 2);
@@ -215,16 +240,16 @@ static int sc_transceive(struct sc_card *card, struct sc_apdu *apdu)
 	if (recvsize > 0) {
 		memcpy(apdu->resp, rbuf, data_bytes);
 		if (apdu->sensitive)
-			memset(rbuf, 0, recvsize);
+			sc_mem_clear(rbuf, recvsize);
 	}
 	return 0;
 }
 
-int sc_transmit_apdu(struct sc_card *card, struct sc_apdu *apdu)
+int sc_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu)
 {
 	int r;
 	size_t orig_resplen;
-	
+
 	assert(card != NULL && apdu != NULL);
 	SC_FUNC_CALLED(card->ctx, 4);
 	orig_resplen = apdu->resplen;
@@ -251,6 +276,9 @@ int sc_transmit_apdu(struct sc_card *card, struct sc_apdu *apdu)
 	if (apdu->sw1 == 0x6C && apdu->resplen == 0) {
 		apdu->resplen = orig_resplen;
 		apdu->le = apdu->sw2;
+		/* Fix for cards (e.g. belpic) that need a delay on fast readers */
+		if (card->wait_resend_apdu != 0)
+			msleep(card->wait_resend_apdu);
 		r = sc_transceive(card, apdu);
 		if (r != 0) {
 			sc_unlock(card);
@@ -258,7 +286,7 @@ int sc_transmit_apdu(struct sc_card *card, struct sc_apdu *apdu)
 		}
 	}
 	if (apdu->sw1 == 0x61 && apdu->resplen == 0) {
-		unsigned int le;
+		size_t le;
 
 		if (orig_resplen == 0) {
 			apdu->sw1 = 0x90;	/* FIXME: should we do this? */
@@ -274,18 +302,16 @@ int sc_transmit_apdu(struct sc_card *card, struct sc_apdu *apdu)
 			SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 		}
 		r = card->ops->get_response(card, apdu, le);
-		sc_unlock(card);
 		if (r < 0) {
+			sc_unlock(card);
 			SC_FUNC_RETURN(card->ctx, 2, r);
 		}
-		else
-			return 0;
 	}
 	sc_unlock(card);
 	return 0;
 }
 
-void sc_format_apdu(struct sc_card *card, struct sc_apdu *apdu,
+void sc_format_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		    int cse, int ins, int p1, int p2)
 {
 	assert(card != NULL && apdu != NULL);
@@ -299,19 +325,19 @@ void sc_format_apdu(struct sc_card *card, struct sc_apdu *apdu,
 	return;
 }
 
-static struct sc_card * sc_card_new(void)
+static sc_card_t * sc_card_new(void)
 {
-	struct sc_card *card;
-	
-	card = (struct sc_card *) malloc(sizeof(struct sc_card));
+	sc_card_t *card;
+
+	card = (sc_card_t *) calloc(1, sizeof(struct sc_card));
 	if (card == NULL)
 		return NULL;
-	memset(card, 0, sizeof(struct sc_card));
 	card->ops = (struct sc_card_operations *) malloc(sizeof(struct sc_card_operations));
 	if (card->ops == NULL) {
 		free(card);
 		return NULL;
 	}
+	card->type = -1;
 	card->app_count = -1;
 	card->magic = SC_CARD_MAGIC;
 	card->mutex = sc_mutex_new();
@@ -319,7 +345,7 @@ static struct sc_card * sc_card_new(void)
 	return card;
 }
 
-static void sc_card_free(struct sc_card *card)
+static void sc_card_free(sc_card_t *card)
 {
 	assert(sc_card_valid(card));
 	sc_free_apps(card);
@@ -329,18 +355,18 @@ static void sc_card_free(struct sc_card *card)
 	if (card->algorithms != NULL)
 		free(card->algorithms);
 	sc_mutex_free(card->mutex);
-	memset(card, 0, sizeof(*card));
+	sc_mem_clear(card, sizeof(*card));
 	free(card);
 }
 
-int sc_connect_card(struct sc_reader *reader, int slot_id,
-		    struct sc_card **card_out)
+int sc_connect_card(sc_reader_t *reader, int slot_id,
+		    sc_card_t **card_out)
 {
-	struct sc_card *card;
-	struct sc_context *ctx = reader->ctx;
-	struct sc_slot_info *slot = _sc_get_slot_info(reader, slot_id);
+	sc_card_t *card;
+	sc_context_t *ctx = reader->ctx;
+	sc_slot_info_t *slot = _sc_get_slot_info(reader, slot_id);
 	struct sc_card_driver *driver;
-	int i, r = 0, connected = 0;
+	int i, r = 0, idx;
 
 	assert(card_out != NULL);
 	SC_FUNC_CALLED(ctx, 1);
@@ -355,7 +381,6 @@ int sc_connect_card(struct sc_reader *reader, int slot_id,
 	r = reader->ops->connect(reader, slot);
 	if (r)
 		goto err;
-	connected = 1;
 
 	card->reader = reader;
 	card->slot = slot;
@@ -372,10 +397,30 @@ int sc_connect_card(struct sc_reader *reader, int slot_id,
 
 	/* See if the ATR matches any ATR specified in the config file */
 	if ((driver = ctx->forced_driver) == NULL) {
+		if (ctx->debug >= 3)
+			sc_debug(ctx, "matching configured ATRs\n");
 		for (i = 0; ctx->card_drivers[i] != NULL; i++) {
 			driver = ctx->card_drivers[i];
-			if (_sc_match_atr(card, driver->atr_map, NULL) >= 0)
+
+			if (driver->atr_map == NULL ||
+			    !strcmp(driver->short_name, "default")) {
+				driver = NULL;
+				continue;
+			}
+			if (ctx->debug >= 3)
+				sc_debug(ctx, "trying driver: %s\n", driver->short_name);
+			idx = _sc_match_atr(card, driver->atr_map, NULL);
+			if (idx >= 0) {
+				struct sc_atr_table *src = &driver->atr_map[idx];
+
+				if (ctx->debug >= 3)
+					sc_debug(ctx, "matched: %s\n", driver->name);
+				/* It's up to card driver to notice these correctly */
+				card->name = src->name;
+				card->type = src->type;
+				card->flags = src->flags;
 				break;
+			}
 			driver = NULL;
 		}
 	}
@@ -393,31 +438,35 @@ int sc_connect_card(struct sc_reader *reader, int slot_id,
 				goto err;
 			}
 		}
-	} else for (i = 0; ctx->card_drivers[i] != NULL; i++) {
-		struct sc_card_driver *drv = ctx->card_drivers[i];
-		const struct sc_card_operations *ops = drv->ops;
-		
+	} else {
 		if (ctx->debug >= 3)
-			sc_debug(ctx, "trying driver: %s\n", drv->name);
-		if (ops == NULL || ops->match_card == NULL)
-			continue;
-		if (ops->match_card(card) != 1)
-			continue;
-		if (ctx->debug >= 3)
-			sc_debug(ctx, "matched: %s\n", drv->name);
-		memcpy(card->ops, ops, sizeof(struct sc_card_operations));
-		card->driver = drv;
-		r = ops->init(card);
-		if (r) {
-			sc_error(ctx, "driver '%s' init() failed: %s\n", drv->name,
-			      sc_strerror(r));
-			if (r == SC_ERROR_INVALID_CARD) {
-				card->driver = NULL;
+			sc_debug(ctx, "matching built-in ATRs\n");
+		for (i = 0; ctx->card_drivers[i] != NULL; i++) {
+			struct sc_card_driver *drv = ctx->card_drivers[i];
+			const struct sc_card_operations *ops = drv->ops;
+
+			if (ctx->debug >= 3)
+				sc_debug(ctx, "trying driver: %s\n", drv->short_name);
+			if (ops == NULL || ops->match_card == NULL)
 				continue;
+			if (ops->match_card(card) != 1)
+				continue;
+			if (ctx->debug >= 3)
+				sc_debug(ctx, "matched: %s\n", drv->name);
+			memcpy(card->ops, ops, sizeof(struct sc_card_operations));
+			card->driver = drv;
+			r = ops->init(card);
+			if (r) {
+				sc_error(ctx, "driver '%s' init() failed: %s\n", drv->name,
+				      sc_strerror(r));
+				if (r == SC_ERROR_INVALID_CARD) {
+					card->driver = NULL;
+					continue;
+				}
+				goto err;
 			}
-			goto err;
+			break;
 		}
-		break;
 	}
 	if (card->driver == NULL) {
 		sc_error(ctx, "unable to find driver for inserted card\n");
@@ -428,6 +477,7 @@ int sc_connect_card(struct sc_reader *reader, int slot_id,
 		card->name = card->driver->name;
 	*card_out = card;
 
+	sc_debug(ctx, "card info: %s, %i, 0x%X\n", card->name, card->type, card->flags);
 	SC_FUNC_RETURN(ctx, 1, 0);
 err:
 	if (card != NULL)
@@ -435,9 +485,9 @@ err:
 	SC_FUNC_RETURN(ctx, 1, r);
 }
 
-int sc_disconnect_card(struct sc_card *card, int action)
+int sc_disconnect_card(sc_card_t *card, int action)
 {
-	struct sc_context *ctx;
+	sc_context_t *ctx;
 	assert(sc_card_valid(card));
 	ctx = card->ctx;
 	SC_FUNC_CALLED(ctx, 1);
@@ -446,7 +496,7 @@ int sc_disconnect_card(struct sc_card *card, int action)
 		int r = card->ops->finish(card);
 		if (r)
 			sc_error(card->ctx, "card driver finish() failed: %s\n",
-			      sc_strerror(r)); 
+			      sc_strerror(r));
 	}
 	if (card->reader->ops->disconnect) {
 		int r = card->reader->ops->disconnect(card->reader, card->slot, action);
@@ -458,10 +508,10 @@ int sc_disconnect_card(struct sc_card *card, int action)
 	SC_FUNC_RETURN(ctx, 1, 0);
 }
 
-int sc_lock(struct sc_card *card)
+int sc_lock(sc_card_t *card)
 {
 	int r = 0;
-	
+
 	assert(card != NULL);
 	sc_mutex_lock(card->mutex);
 	if (card->lock_count == 0) {
@@ -477,7 +527,7 @@ int sc_lock(struct sc_card *card)
 	return r;
 }
 
-int sc_unlock(struct sc_card *card)
+int sc_unlock(sc_card_t *card)
 {
 	int r = 0;
 
@@ -500,25 +550,25 @@ int sc_unlock(struct sc_card *card)
 	if (card->lock_count == 1) {
 		if (card->reader->ops->unlock != NULL)
 			r = card->reader->ops->unlock(card->reader, card->slot);
-        }
-        card->lock_count--;
-        sc_mutex_unlock(card->mutex);
+	}
+	card->lock_count--;
+	sc_mutex_unlock(card->mutex);
 	return r;
 }
 
-int sc_list_files(struct sc_card *card, u8 *buf, size_t buflen)
+int sc_list_files(sc_card_t *card, u8 *buf, size_t buflen)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 1);
-        if (card->ops->list_files == NULL)
+	if (card->ops->list_files == NULL)
 		SC_FUNC_RETURN(card->ctx, 1, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->list_files(card, buf, buflen);
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
-int sc_create_file(struct sc_card *card, struct sc_file *file)
+int sc_create_file(sc_card_t *card, sc_file_t *file)
 {
 	int r;
 
@@ -530,13 +580,13 @@ int sc_create_file(struct sc_card *card, struct sc_file *file)
 				in_path->type,
 				sc_print_path(in_path), file->size);
 	}
-        if (card->ops->create_file == NULL)
+	if (card->ops->create_file == NULL)
 		SC_FUNC_RETURN(card->ctx, 1, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->create_file(card, file);
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
-int sc_delete_file(struct sc_card *card, const struct sc_path *path)
+int sc_delete_file(sc_card_t *card, const sc_path_t *path)
 {
 	int r;
 
@@ -546,13 +596,13 @@ int sc_delete_file(struct sc_card *card, const struct sc_path *path)
 				path->type,
 				sc_print_path(path));
 	}
-        if (card->ops->delete_file == NULL)
+	if (card->ops->delete_file == NULL)
 		SC_FUNC_RETURN(card->ctx, 1, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->delete_file(card, path);
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
-int sc_read_binary(struct sc_card *card, unsigned int idx,
+int sc_read_binary(sc_card_t *card, unsigned int idx,
 		   unsigned char *buf, size_t count, unsigned long flags)
 {
 	size_t max_le = card->max_recv_size;
@@ -591,10 +641,10 @@ int sc_read_binary(struct sc_card *card, unsigned int idx,
 		SC_FUNC_RETURN(card->ctx, 2, bytes_read);
 	}
 	r = card->ops->read_binary(card, idx, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_write_binary(struct sc_card *card, unsigned int idx,
+int sc_write_binary(sc_card_t *card, unsigned int idx,
 		    const u8 *buf, size_t count, unsigned long flags)
 {
 	size_t max_lc = card->max_send_size;
@@ -633,10 +683,10 @@ int sc_write_binary(struct sc_card *card, unsigned int idx,
 		SC_FUNC_RETURN(card->ctx, 2, bytes_written);
 	}
 	r = card->ops->write_binary(card, idx, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_update_binary(struct sc_card *card, unsigned int idx,
+int sc_update_binary(sc_card_t *card, unsigned int idx,
 		     const u8 *buf, size_t count, unsigned long flags)
 {
 	size_t max_lc = card->max_send_size;
@@ -675,12 +725,12 @@ int sc_update_binary(struct sc_card *card, unsigned int idx,
 		SC_FUNC_RETURN(card->ctx, 2, bytes_written);
 	}
 	r = card->ops->update_binary(card, idx, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_select_file(struct sc_card *card,
-		   const struct sc_path *in_path,
-		   struct sc_file **file)
+int sc_select_file(sc_card_t *card,
+		   const sc_path_t *in_path,
+		   sc_file_t **file)
 {
 	int r;
 
@@ -710,7 +760,7 @@ int sc_select_file(struct sc_card *card,
 	/* Remember file path */
 	if (r == 0 && file && *file)
 		(*file)->path = *in_path;
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
 int sc_get_data(sc_card_t *card, unsigned int tag, u8 *buf, size_t len)
@@ -718,10 +768,10 @@ int sc_get_data(sc_card_t *card, unsigned int tag, u8 *buf, size_t len)
 	int	r;
 
 	sc_debug(card->ctx, "called, tag=%04x\n", tag);
-        if (card->ops->get_data == NULL)
+	if (card->ops->get_data == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->get_data(card, tag, buf, len);
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
 int sc_put_data(sc_card_t *card, unsigned int tag, const u8 *buf, size_t len)
@@ -729,77 +779,77 @@ int sc_put_data(sc_card_t *card, unsigned int tag, const u8 *buf, size_t len)
 	int	r;
 
 	sc_debug(card->ctx, "called, tag=%04x\n", tag);
-        if (card->ops->put_data == NULL)
+	if (card->ops->put_data == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->put_data(card, tag, buf, len);
-        SC_FUNC_RETURN(card->ctx, 1, r);
+	SC_FUNC_RETURN(card->ctx, 1, r);
 }
 
-int sc_get_challenge(struct sc_card *card, u8 *rnd, size_t len)
+int sc_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->get_challenge == NULL)
+	if (card->ops->get_challenge == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->get_challenge(card, rnd, len);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_read_record(struct sc_card *card, unsigned int rec_nr, u8 *buf,
+int sc_read_record(sc_card_t *card, unsigned int rec_nr, u8 *buf,
 		   size_t count, unsigned long flags)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->read_record == NULL)
+	if (card->ops->read_record == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->read_record(card, rec_nr, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_write_record(struct sc_card *card, unsigned int rec_nr, const u8 * buf,
+int sc_write_record(sc_card_t *card, unsigned int rec_nr, const u8 * buf,
 		    size_t count, unsigned long flags)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->write_record == NULL)
+	if (card->ops->write_record == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->write_record(card, rec_nr, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_append_record(struct sc_card *card, const u8 * buf, size_t count,
+int sc_append_record(sc_card_t *card, const u8 * buf, size_t count,
 		     unsigned long flags)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->append_record == NULL)
+	if (card->ops->append_record == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->append_record(card, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_update_record(struct sc_card *card, unsigned int rec_nr, const u8 * buf,
+int sc_update_record(sc_card_t *card, unsigned int rec_nr, const u8 * buf,
 		     size_t count, unsigned long flags)
 {
 	int r;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->update_record == NULL)
+	if (card->ops->update_record == NULL)
 		SC_FUNC_RETURN(card->ctx, 2, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->update_record(card, rec_nr, buf, count, flags);
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int sc_delete_record(struct sc_card *card, unsigned int rec_nr)
+int sc_delete_record(sc_card_t *card, unsigned int rec_nr)
 {
 	int r;
 
@@ -811,7 +861,7 @@ int sc_delete_record(struct sc_card *card, unsigned int rec_nr)
 	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-inline int sc_card_valid(const struct sc_card *card) {
+inline int sc_card_valid(const sc_card_t *card) {
 #ifndef NDEBUG
 	assert(card != NULL);
 #endif
@@ -819,13 +869,13 @@ inline int sc_card_valid(const struct sc_card *card) {
 }
 
 int
-sc_card_ctl(struct sc_card *card, unsigned long cmd, void *args)
+sc_card_ctl(sc_card_t *card, unsigned long cmd, void *args)
 {
 	int r = SC_ERROR_NOT_SUPPORTED;
 
 	assert(card != NULL);
 	SC_FUNC_CALLED(card->ctx, 2);
-        if (card->ops->card_ctl != NULL)
+	if (card->ops->card_ctl != NULL)
 		r = card->ops->card_ctl(card, cmd, args);
 
 	/* suppress "not supported" error messages */
@@ -833,15 +883,15 @@ sc_card_ctl(struct sc_card *card, unsigned long cmd, void *args)
 		sc_debug(card->ctx, "card_ctl(%lu) not supported\n", cmd);
 		return r;
 	}
-        SC_FUNC_RETURN(card->ctx, 2, r);
+	SC_FUNC_RETURN(card->ctx, 2, r);
 }
 
-int _sc_card_add_algorithm(struct sc_card *card, const struct sc_algorithm_info *info)
+int _sc_card_add_algorithm(sc_card_t *card, const sc_algorithm_info_t *info)
 {
-	struct sc_algorithm_info *p;
-	
+	sc_algorithm_info_t *p;
+
 	assert(sc_card_valid(card) && info != NULL);
-	p = (struct sc_algorithm_info *) realloc(card->algorithms, (card->algorithm_count + 1) * sizeof(*info));
+	p = (sc_algorithm_info_t *) realloc(card->algorithms, (card->algorithm_count + 1) * sizeof(*info));
 	if (!p) {
 		if (card->algorithms)
 			free(card->algorithms);
@@ -856,28 +906,28 @@ int _sc_card_add_algorithm(struct sc_card *card, const struct sc_algorithm_info
 	return 0;
 }
 
-int _sc_card_add_rsa_alg(struct sc_card *card, unsigned int key_length,
+int _sc_card_add_rsa_alg(sc_card_t *card, unsigned int key_length,
 			 unsigned long flags, unsigned long exponent)
 {
-	struct sc_algorithm_info info;
+	sc_algorithm_info_t info;
 
 	memset(&info, 0, sizeof(info));
 	info.algorithm = SC_ALGORITHM_RSA;
 	info.key_length = key_length;
 	info.flags = flags;
 	info.u._rsa.exponent = exponent;
-	
+
 	return _sc_card_add_algorithm(card, &info);
 }
 
-struct sc_algorithm_info * _sc_card_find_rsa_alg(struct sc_card *card,
+sc_algorithm_info_t * _sc_card_find_rsa_alg(sc_card_t *card,
 						 unsigned int key_length)
 {
 	int i;
 
 	for (i = 0; i < card->algorithm_count; i++) {
-		struct sc_algorithm_info *info = &card->algorithms[i];
-		
+		sc_algorithm_info_t *info = &card->algorithms[i];
+
 		if (info->algorithm != SC_ALGORITHM_RSA)
 			continue;
 		if (info->key_length != key_length)
@@ -887,45 +937,208 @@ struct sc_algorithm_info * _sc_card_find_rsa_alg(struct sc_card *card,
 	return NULL;
 }
 
-int _sc_match_atr(struct sc_card *card, struct sc_atr_table *table, int *id_out)
+static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, u8 *atr, size_t atr_len)
 {
-	const u8 *atr = card->atr;
-	size_t atr_len = card->atr_len;
-	int i = 0;
-	
-	if (table == NULL)
+	u8 *card_atr_bin = atr;
+	size_t card_atr_bin_len = atr_len;
+	char card_atr_hex[3 * SC_MAX_ATR_SIZE];
+	size_t card_atr_hex_len;
+	unsigned int i = 0;
+
+	if (ctx == NULL || table == NULL || atr == NULL)
 		return -1;
+	sc_bin_to_hex(card_atr_bin, card_atr_bin_len, card_atr_hex, sizeof(card_atr_hex), ':');
+	card_atr_hex_len = strlen(card_atr_hex);
+
+	if (ctx->debug >= 4)
+		sc_debug(ctx, "ATR     : %s\n", card_atr_hex);
 
 	for (i = 0; table[i].atr != NULL; i++) {
-		if (table[i].atr_len != atr_len)
+		const char *tatr = table[i].atr;
+		const char *matr = table[i].atrmask;
+		size_t tatr_len = strlen(tatr);
+
+		if (ctx->debug >= 4)
+			sc_debug(ctx, "ATR try : %s\n", tatr);
+
+		if (tatr_len != card_atr_hex_len)
 			continue;
-		if (memcmp(table[i].atr, atr, atr_len) != 0)
-			continue;
-		if (id_out != NULL)
-			*id_out = table[i].id;
+		if (matr != NULL) {
+			u8 mbin[SC_MAX_ATR_SIZE], tbin[SC_MAX_ATR_SIZE];
+			size_t mbin_len, tbin_len, s, matr_len;
+
+			if (ctx->debug >= 4)
+				sc_debug(ctx, "ATR mask: %s\n", matr);
+
+			matr_len = strlen(matr);
+			if (tatr_len != matr_len)
+				continue;
+			mbin_len = sizeof(mbin);
+			sc_hex_to_bin(matr, mbin, &mbin_len);
+			if (mbin_len != card_atr_bin_len)
+				continue;
+			for (s = 0; s < mbin_len; s++)
+				mbin[s] = (card_atr_bin[s] & mbin[s]);
+			tbin_len = sizeof(tbin);
+			sc_hex_to_bin(tatr, tbin, &tbin_len);
+			if (memcmp(tbin, mbin, tbin_len) != 0)
+				continue;
+		} else {
+			if (strncasecmp(tatr, card_atr_hex, tatr_len) != 0)
+				continue;
+		}
 		return i;
 	}
 	return -1;
 }
 
-int _sc_add_atr(struct sc_card_driver *driver,
-		 const u8 *atr, size_t atrlen, int id)
+int _sc_match_atr(sc_card_t *card, struct sc_atr_table *table, int *type_out)
 {
-	struct sc_atr_table *map, *entry;
-	u8 *dst_atr;
+	int res;
+
+	if (card == NULL)
+		return -1;
+	res = match_atr_table(card->ctx, table, card->atr, card->atr_len);
+	if (res < 0)
+		return res;
+	if (type_out != NULL)
+		*type_out = table[res].type;
+	return res;
+}
+
+scconf_block *_sc_match_atr_block(sc_context_t *ctx, struct sc_card_driver *driver, u8 *atr, size_t atr_len)
+{
+	struct sc_card_driver *drv;
+	struct sc_atr_table *table;
+	int res;
+
+	if (ctx == NULL)
+		return NULL;
+	if (driver) {
+		drv = driver;
+		table = drv->atr_map;
+		res = match_atr_table(ctx, table, atr, atr_len);
+		if (res < 0)
+			return NULL;
+		return table[res].card_atr;
+	} else {
+		unsigned int i;
+
+		for (i = 0; ctx->card_drivers[i] != NULL; i++) {
+			drv = ctx->card_drivers[i];
+			table = drv->atr_map;
+			res = match_atr_table(ctx, table, atr, atr_len);
+			if (res < 0)
+				continue;
+			return table[res].card_atr;
+		}
+	}
+	return NULL;
+}
+
+int _sc_add_atr(sc_context_t *ctx, struct sc_card_driver *driver, struct sc_atr_table *src)
+{
+	struct sc_atr_table *map, *dst;
 
 	map = (struct sc_atr_table *) realloc(driver->atr_map,
 			(driver->natrs + 2) * sizeof(struct sc_atr_table));
 	if (!map)
 		return SC_ERROR_OUT_OF_MEMORY;
 	driver->atr_map = map;
-	if (!(dst_atr = (u8 *) malloc(atrlen)))
+	dst = &driver->atr_map[driver->natrs++];
+	memset(dst, 0, sizeof(*dst));
+	memset(&driver->atr_map[driver->natrs], 0, sizeof(struct sc_atr_table));
+	dst->atr = strdup(src->atr);
+	if (!dst->atr)
 		return SC_ERROR_OUT_OF_MEMORY;
-	entry = &driver->atr_map[driver->natrs++];
-	memset(entry+1, 0, sizeof(*entry));
-	entry->atr = dst_atr;
-	entry->atr_len = atrlen;
-	entry->id = id;
-	memcpy(dst_atr, atr, atrlen);
-	return 0;
+	if (src->atrmask) {
+		dst->atrmask = strdup(src->atrmask);
+		if (!dst->atrmask)
+			return SC_ERROR_OUT_OF_MEMORY;
+	} else {
+		dst->atrmask = NULL;
+	}
+	if (src->name) {
+		dst->name = strdup(src->name);
+		if (!dst->name)
+			return SC_ERROR_OUT_OF_MEMORY;
+	} else {
+		dst->name = NULL;
+	}
+	dst->type = src->type;
+	dst->flags = src->flags;
+	dst->card_atr = src->card_atr;
+	return SC_SUCCESS;
 }
+
+int _sc_free_atr(sc_context_t *ctx, struct sc_card_driver *driver)
+{
+	unsigned int i;
+
+	for (i = 0; i < driver->natrs; i++) {
+		struct sc_atr_table *src = &driver->atr_map[i];
+
+		if (src->atr)
+			free(src->atr);
+		if (src->atrmask)
+			free(src->atrmask);
+		if (src->name)
+			free(src->name);
+		src->card_atr = NULL;
+		src = NULL;
+	}
+	if (driver->atr_map)
+		free(driver->atr_map);
+	driver->atr_map = NULL;
+	driver->natrs = 0;
+
+	return SC_SUCCESS;
+}
+
+int _sc_check_forced_protocol(sc_context_t *ctx, u8 *atr, size_t atr_len, unsigned int *protocol)
+{
+	scconf_block *atrblock = NULL;
+	int ok = 0;
+
+	if (!protocol)
+		return 0;
+	atrblock = _sc_match_atr_block(ctx, NULL, atr, atr_len);
+	if (atrblock != NULL) {
+		const char *forcestr;
+
+		forcestr = scconf_get_str(atrblock, "force_protocol", "unknown");
+		if (!strcmp(forcestr, "t0")) {
+			*protocol = SC_PROTO_T0;
+			ok = 1;
+		} else if (!strcmp(forcestr, "t1")) {
+			*protocol = SC_PROTO_T1;
+			ok = 1;
+		} else if (!strcmp(forcestr, "raw")) {
+			*protocol = SC_PROTO_RAW;
+			ok = 1;
+		}
+		if (ok)
+			sc_debug(ctx, "force_protocol: %s\n", forcestr);
+	}
+	return ok;
+}
+
+scconf_block *_get_conf_block(sc_context_t *ctx, const char *name1, const char *name2, int priority)
+{
+	int i;
+	scconf_block *conf_block = NULL;
+	
+	for (i = 0; ctx->conf_blocks[i] != NULL; i++) {
+		scconf_block **blocks;
+		
+		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i], name1, name2);
+		if (!blocks)
+			return NULL;
+		conf_block = blocks[0];
+		free(blocks);
+		if (conf_block != NULL && priority)
+			break;
+	}
+	return conf_block;
+}
+
