@@ -27,57 +27,57 @@
 #include <assert.h>
 
 static const struct sc_asn1_entry c_asn1_com_key_attr[] = {
-	{ "iD",		 SC_ASN1_PKCS15_ID, ASN1_OCTET_STRING, 0, NULL },
-	{ "usage",	 SC_ASN1_BIT_FIELD, ASN1_BIT_STRING, 0, NULL },
-	{ "native",	 SC_ASN1_BOOLEAN, ASN1_BOOLEAN, SC_ASN1_OPTIONAL, NULL },
-	{ "accessFlags", SC_ASN1_BIT_FIELD, ASN1_BIT_STRING, SC_ASN1_OPTIONAL, NULL },
-	{ "keyReference",SC_ASN1_INTEGER, ASN1_INTEGER, SC_ASN1_OPTIONAL, NULL },
-	{ NULL }
+	{ "iD",		 SC_ASN1_PKCS15_ID, ASN1_OCTET_STRING, 0, NULL, NULL },
+	{ "usage",	 SC_ASN1_BIT_FIELD, ASN1_BIT_STRING, 0, NULL, NULL },
+	{ "native",	 SC_ASN1_BOOLEAN, ASN1_BOOLEAN, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "accessFlags", SC_ASN1_BIT_FIELD, ASN1_BIT_STRING, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "keyReference",SC_ASN1_INTEGER, ASN1_INTEGER, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_com_pubkey_attr[] = {
         /* FIXME */
-	{ NULL }
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_rsakey_attr[] = {
-	{ "value",	   SC_ASN1_PATH, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
-	{ "modulusLength", SC_ASN1_INTEGER, ASN1_INTEGER, 0, NULL },
-	{ "keyInfo",	   SC_ASN1_INTEGER, ASN1_INTEGER, SC_ASN1_OPTIONAL, NULL },
-	{ NULL }
+	{ "value",	   SC_ASN1_PATH, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ "modulusLength", SC_ASN1_INTEGER, ASN1_INTEGER, 0, NULL, NULL },
+	{ "keyInfo",	   SC_ASN1_INTEGER, ASN1_INTEGER, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_rsa_type_attr[] = {
-	{ "publicRSAKeyAttributes", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
-	{ NULL }
+	{ "publicRSAKeyAttributes", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_dsakey_attr[] = {
-	{ "value",	   SC_ASN1_PATH, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
-	{ NULL }
+	{ "value",	   SC_ASN1_PATH, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_dsa_type_attr[] = {
-	{ "publicDSAKeyAttributes", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
-	{ NULL }
+	{ "publicDSAKeyAttributes", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_pubkey_choice[] = {
-	{ "publicRSAKey", SC_ASN1_PKCS15_OBJECT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL },
-	{ "publicDSAKey", SC_ASN1_PKCS15_OBJECT, 2 | SC_ASN1_CTX | SC_ASN1_CONS, 0, NULL },
-	{ NULL }
+	{ "publicRSAKey", SC_ASN1_PKCS15_OBJECT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ "publicDSAKey", SC_ASN1_PKCS15_OBJECT, 2 | SC_ASN1_CTX | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static const struct sc_asn1_entry c_asn1_pubkey[] = {
-	{ "publicKey",	SC_ASN1_CHOICE, },
-	{ NULL }
+	{ "publicKey",	SC_ASN1_CHOICE, 0, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 				 struct sc_pkcs15_object *obj,
 				 const u8 ** buf, size_t *buflen)
 {
-        struct sc_context *ctx = p15card->card->ctx;
+        sc_context_t *ctx = p15card->card->ctx;
         struct sc_pkcs15_pubkey_info info;
 	int r;
 	size_t usage_len = sizeof(info.usage);
@@ -142,7 +142,7 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 
 	return 0;
 }
-int sc_pkcs15_encode_pukdf_entry(struct sc_context *ctx,
+int sc_pkcs15_encode_pukdf_entry(sc_context_t *ctx,
 				 const struct sc_pkcs15_object *obj,
 				 u8 **buf, size_t *buflen)
 {
@@ -213,26 +213,26 @@ int sc_pkcs15_encode_pukdf_entry(struct sc_context *ctx,
 }
 
 static struct sc_asn1_entry c_asn1_public_key[2] = {
-	{ "publicKeyCoefficients", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, },
-	{ NULL }
+	{ "publicKeyCoefficients", SC_ASN1_STRUCT, ASN1_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static struct sc_asn1_entry c_asn1_rsa_pub_coefficients[3] = {
-	{ "modulus",  SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ "exponent", SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ NULL }
+	{ "modulus",  SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ "exponent", SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
 static struct sc_asn1_entry c_asn1_dsa_pub_coefficients[5] = {
-	{ "publicKey",SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ "paramP",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ "paramQ",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ "paramG",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, },
-	{ NULL },
+	{ "publicKey",SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ "paramP",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ "paramQ",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ "paramG",   SC_ASN1_OCTET_STRING, ASN1_INTEGER, SC_ASN1_ALLOC|SC_ASN1_UNSIGNED, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL },
 };
 
 int
-sc_pkcs15_decode_pubkey_rsa(struct sc_context *ctx,
+sc_pkcs15_decode_pubkey_rsa(sc_context_t *ctx,
 			struct sc_pkcs15_pubkey_rsa *key,
 			const u8 *buf, size_t buflen)
 {
@@ -256,7 +256,7 @@ sc_pkcs15_decode_pubkey_rsa(struct sc_context *ctx,
 }
 
 int
-sc_pkcs15_encode_pubkey_rsa(struct sc_context *ctx,
+sc_pkcs15_encode_pubkey_rsa(sc_context_t *ctx,
 		struct sc_pkcs15_pubkey_rsa *key,
 		u8 **buf, size_t *buflen)
 {
@@ -280,7 +280,7 @@ sc_pkcs15_encode_pubkey_rsa(struct sc_context *ctx,
 }
 
 int
-sc_pkcs15_decode_pubkey_dsa(struct sc_context *ctx,
+sc_pkcs15_decode_pubkey_dsa(sc_context_t *ctx,
 		struct sc_pkcs15_pubkey_dsa *key,
 		const u8 *buf, size_t buflen)
 {
@@ -309,7 +309,7 @@ sc_pkcs15_decode_pubkey_dsa(struct sc_context *ctx,
 }
 
 int
-sc_pkcs15_encode_pubkey_dsa(struct sc_context *ctx,
+sc_pkcs15_encode_pubkey_dsa(sc_context_t *ctx,
 		struct sc_pkcs15_pubkey_dsa *key,
 		u8 **buf, size_t *buflen)
 {
@@ -337,7 +337,7 @@ sc_pkcs15_encode_pubkey_dsa(struct sc_context *ctx,
 }
 
 int
-sc_pkcs15_encode_pubkey(struct sc_context *ctx,
+sc_pkcs15_encode_pubkey(sc_context_t *ctx,
 		struct sc_pkcs15_pubkey *key,
 		u8 **buf, size_t *len)
 {
@@ -351,7 +351,7 @@ sc_pkcs15_encode_pubkey(struct sc_context *ctx,
 }
 
 int
-sc_pkcs15_decode_pubkey(struct sc_context *ctx,
+sc_pkcs15_decode_pubkey(sc_context_t *ctx,
 		struct sc_pkcs15_pubkey *key,
 		const u8 *buf, size_t len)
 {
@@ -433,7 +433,7 @@ void sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *key)
 		break;
 	}
 	free(key->data.value);
-	memset(key, 0, sizeof(*key));
+	sc_mem_clear(key, sizeof(*key));
 }
 
 void sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *key)
@@ -441,3 +441,10 @@ void sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *key)
 	sc_pkcs15_erase_pubkey(key);
 	free(key);
 }
+
+void sc_pkcs15_free_pubkey_info(sc_pkcs15_pubkey_info_t *key)
+{
+	if (key->subject)
+		free(key->subject);
+	free(key);
+}
