@@ -18,8 +18,8 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include "internal.h"
-#include "pkcs15.h"
+#include <opensc/pkcs15.h>
+#include <opensc/cardctl.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
@@ -27,6 +27,8 @@
 #define MANU_ID		"Giesecke & Devrient GmbH"
 #define STARCERT	"StarCertV2201"
 
+int sc_pkcs15emu_starcert_init_ex(sc_pkcs15_card_t *, sc_pkcs15emu_opt_t *);
+
 typedef struct cdata_st {
 	const char *label;
 	int	    authority;
@@ -35,14 +37,6 @@ typedef struct cdata_st {
 	int         obj_flags;
 } cdata;
 
-const cdata certs[] = {
-	{"DS certificate", 0, "3F00DF01C000","1", SC_PKCS15_CO_FLAG_MODIFIABLE},
-	{"CA certificate", 1, "3F00DF01C008","2", SC_PKCS15_CO_FLAG_MODIFIABLE},
-	{"KE certificate", 0, "3F00DF01C200","3", SC_PKCS15_CO_FLAG_MODIFIABLE},
-	{"AUT certificate",0, "3F00DF01C500","4", SC_PKCS15_CO_FLAG_MODIFIABLE},
-	{NULL, 0, NULL, 0, 0}
-};
-
 typedef struct pdata_st {
 	const char *id;
 	const char *label;
@@ -51,20 +45,13 @@ typedef struct pdata_st {
 	int         type;
 	unsigned int maxlen;
 	unsigned int minlen;
+	unsigned int storedlen;
 	int         flags;	
 	int         tries_left;
 	const char  pad_char;
 	int         obj_flags;
 } pindata; 
 
-const pindata pins[] = {
-	{ "99", "DS pin", "3F00DF01", 0x99, SC_PKCS15_PIN_TYPE_ASCII_NUMERIC,
-	  8, 8, SC_PKCS15_PIN_FLAG_NEEDS_PADDING | SC_PKCS15_PIN_FLAG_LOCAL, 
-	  3, 0x00, 
-	  SC_PKCS15_CO_FLAG_MODIFIABLE | SC_PKCS15_CO_FLAG_PRIVATE },
-	{ NULL, NULL, NULL, 0, 0, 0, 0, 0, 0, 0}
-};
-
 typedef struct prdata_st {
 	const char *id;
 	const char *label;
@@ -87,17 +74,7 @@ typedef struct prdata_st {
 			SC_PKCS15_PRKEY_USAGE_UNWRAP  | \
 			SC_PKCS15_PRKEY_USAGE_SIGN
 
-const prdata prkeys[] = {
-	{ "1", "DS key", 1024, USAGE_NONREP, "3F00DF01",
-	  0x84, "99", SC_PKCS15_CO_FLAG_PRIVATE},
-	{ "3", "KE key", 1024, USAGE_KE, "3F00DF01",
-	  0x85, NULL, SC_PKCS15_CO_FLAG_PRIVATE},
-	{ "4", "AUT key", 1024, USAGE_AUT, "3F00DF01",
-	  0x82, NULL, SC_PKCS15_CO_FLAG_PRIVATE},
-	{ NULL, NULL, 0, 0, NULL, 0, NULL, 0}
-};
-
-int get_cert_len(sc_card_t *card, sc_path_t *path)
+static int get_cert_len(sc_card_t *card, sc_path_t *path)
 {
 	int r;
 	u8  buf[8];
@@ -141,35 +118,54 @@ static int starcert_detect_card(sc_pkcs15_card_t *p15card)
 	return SC_SUCCESS;
 }
 
-int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
+static int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 {
-	int    r, i;
-	struct sc_apdu apdu;
-	u8     rbuf[SC_MAX_APDU_BUFFER_SIZE], buf[256];
-	struct sc_path path;
-	struct sc_file *file = NULL;
-	struct sc_card *card = p15card->card;
+	const cdata certs[] = {
+		{"DS certificate", 0, "3F00DF01C000","1",
+			SC_PKCS15_CO_FLAG_MODIFIABLE},
+		{"CA certificate", 1, "3F00DF01C008","2",
+			SC_PKCS15_CO_FLAG_MODIFIABLE},
+		{"KE certificate", 0, "3F00DF01C200","3",
+			SC_PKCS15_CO_FLAG_MODIFIABLE},
+		{"AUT certificate",0, "3F00DF01C500","4",
+			SC_PKCS15_CO_FLAG_MODIFIABLE},
+		{NULL, 0, NULL, 0, 0}
+	};
 
-	/* use Starcos command GET CARD DATA to determine the
-	 * serial number of the card.
-	 */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xf6, 0x00, 0x00);
-	apdu.cla |= 0x80;
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf);
-	apdu.le   = 256;
-	apdu.lc   = 0;
-	apdu.datalen = 0;
-	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, r, "APDU transmit failed");
-	if (apdu.sw1 != 0x90 || apdu.sw2 != 0x00)
-		return SC_ERROR_INTERNAL;
-	r = sc_bin_to_hex(apdu.resp, apdu.resplen, buf, sizeof(buf), 0);
+	const pindata pins[] = {
+		{ "99", "DS pin", "3F00DF01", 0x99,
+		  SC_PKCS15_PIN_TYPE_ASCII_NUMERIC,
+		  8, 8, 8, SC_PKCS15_PIN_FLAG_NEEDS_PADDING |
+		  SC_PKCS15_PIN_FLAG_LOCAL, -1, 0x00,
+		  SC_PKCS15_CO_FLAG_MODIFIABLE | SC_PKCS15_CO_FLAG_PRIVATE },
+		{ NULL, NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0}
+	};
+
+	const prdata prkeys[] = {
+		{ "1", "DS key", 1024, USAGE_NONREP, "3F00DF01",
+		  0x84, "99", SC_PKCS15_CO_FLAG_PRIVATE},
+		{ "3", "KE key", 1024, USAGE_KE, "3F00DF01",
+		  0x85, NULL, SC_PKCS15_CO_FLAG_PRIVATE},
+		{ "4", "AUT key", 1024, USAGE_AUT, "3F00DF01",
+		  0x82, NULL, SC_PKCS15_CO_FLAG_PRIVATE},
+		{ NULL, NULL, 0, 0, NULL, 0, NULL, 0}
+	};
+
+	int    r, i;
+	char   buf[256];
+	sc_path_t path;
+	sc_file_t *file = NULL;
+	sc_card_t *card = p15card->card;
+	sc_serial_number_t serial;
+
+	/* get serial number */
+	r = sc_card_ctl(card, SC_CARDCTL_GET_SERIALNR, &serial);
+	r = sc_bin_to_hex(serial.value, serial.len, buf, sizeof(buf), 0);
 	if (r != SC_SUCCESS)
 		return SC_ERROR_INTERNAL;
 	if (p15card->serial_number)
 		free(p15card->serial_number);
-	p15card->serial_number = malloc(strlen(buf) + 1);
+	p15card->serial_number = (char *) malloc(strlen(buf) + 1);
 	if (!p15card->serial_number)
 		return SC_ERROR_INTERNAL;
 	strcpy(p15card->serial_number, buf);
@@ -178,52 +174,82 @@ int sc_pkcs15emu_starcert_init(sc_pkcs15_card_t *p15card)
 	/* the manufacturer ID, in this case Giesecke & Devrient GmbH */
 	if (p15card->manufacturer_id)
 		free(p15card->manufacturer_id);
-	p15card->manufacturer_id = malloc(strlen(MANU_ID) + 1);
+	p15card->manufacturer_id = (char *) malloc(strlen(MANU_ID) + 1);
 	if (!p15card->manufacturer_id)
 		return SC_ERROR_INTERNAL;
 	strcpy(p15card->manufacturer_id, MANU_ID);
 
 	/* set certs */
 	for (i = 0; certs[i].label; i++) {
-		struct sc_pkcs15_id  p15Id;
+		struct sc_pkcs15_cert_info cert_info;
+		struct sc_pkcs15_object    cert_obj;
 
-		sc_format_path(certs[i].path, &path);
-		if (!get_cert_len(card, &path))
+		memset(&cert_info, 0, sizeof(cert_info));
+		memset(&cert_obj,  0, sizeof(cert_obj));
+
+		sc_pkcs15_format_id(certs[i].id, &cert_info.id);
+		cert_info.authority = certs[i].authority;
+		sc_format_path(certs[i].path, &cert_info.path);
+		if (!get_cert_len(card, &cert_info.path))
 			/* skip errors */
 			continue;
-		sc_pkcs15_format_id(certs[i].id, &p15Id);
-		sc_pkcs15emu_add_cert(p15card, SC_PKCS15_TYPE_CERT_X509,
-				certs[i].authority, &path, &p15Id,
-				certs[i].label, certs[i].obj_flags);
+
+		strncpy(cert_obj.label, certs[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
+		cert_obj.flags = certs[i].obj_flags;
+
+		r = sc_pkcs15emu_add_x509_cert(p15card, &cert_obj, &cert_info);
+		if (r < 0)
+			return SC_ERROR_INTERNAL;
 	}
 	/* set pins */
 	for (i = 0; pins[i].label; i++) {
-		struct sc_pkcs15_id  p15Id;
+		struct sc_pkcs15_pin_info pin_info;
+		struct sc_pkcs15_object   pin_obj;
 
-		sc_format_path(pins[i].path, &path);
-		sc_pkcs15_format_id(pins[i].id, &p15Id);
-		sc_pkcs15emu_add_pin(p15card, &p15Id, pins[i].label,
-				&path, pins[i].ref, pins[i].type,
-				pins[i].minlen, pins[i].maxlen, pins[i].flags,
-				pins[i].tries_left, pins[i].pad_char,
-				pins[i].obj_flags);
+		memset(&pin_info, 0, sizeof(pin_info));
+		memset(&pin_obj,  0, sizeof(pin_obj));
+
+		sc_pkcs15_format_id(pins[i].id, &pin_info.auth_id);
+		pin_info.reference     = pins[i].ref;
+		pin_info.flags         = pins[i].flags;
+		pin_info.type          = pins[i].type;
+		pin_info.min_length    = pins[i].minlen;
+		pin_info.stored_length = pins[i].storedlen;
+		pin_info.max_length    = pins[i].maxlen;
+		pin_info.pad_char      = pins[i].pad_char;
+		sc_format_path(pins[i].path, &pin_info.path);
+		pin_info.tries_left    = -1;
+
+		strncpy(pin_obj.label, pins[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
+		pin_obj.flags = pins[i].obj_flags;
+
+		r = sc_pkcs15emu_add_pin_obj(p15card, &pin_obj, &pin_info);
+		if (r < 0)
+			return SC_ERROR_INTERNAL;
 	}
 	/* set private keys */
 	for (i = 0; prkeys[i].label; i++) {
-		struct sc_pkcs15_id p15Id, 
-				    authId, *pauthId;
-		sc_format_path(prkeys[i].path, &path);
-		sc_pkcs15_format_id(prkeys[i].id, &p15Id);
-		if (prkeys[i].auth_id) {
-			sc_pkcs15_format_id(prkeys[i].auth_id, &authId);
-			pauthId = &authId;
-		} else	
-			pauthId = NULL;
-		sc_pkcs15emu_add_prkey(p15card, &p15Id, prkeys[i].label,
-				SC_PKCS15_TYPE_PRKEY_RSA,
-				prkeys[i].modulus_len, prkeys[i].usage,
-				&path, prkeys[i].ref, pauthId,
-				prkeys[i].obj_flags);
+		struct sc_pkcs15_prkey_info prkey_info;
+		struct sc_pkcs15_object     prkey_obj;
+
+		memset(&prkey_info, 0, sizeof(prkey_info));
+		memset(&prkey_obj,  0, sizeof(prkey_obj));
+
+		sc_pkcs15_format_id(prkeys[i].id, &prkey_info.id);
+		prkey_info.usage         = prkeys[i].usage;
+		prkey_info.native        = 1;
+		prkey_info.key_reference = prkeys[i].ref;
+		prkey_info.modulus_length= prkeys[i].modulus_len;
+		sc_format_path(prkeys[i].path, &prkey_info.path);
+
+		strncpy(prkey_obj.label, prkeys[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
+		prkey_obj.flags = prkeys[i].obj_flags;
+		if (prkeys[i].auth_id)
+			sc_pkcs15_format_id(prkeys[i].auth_id, &prkey_obj.auth_id);
+
+		r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
+		if (r < 0)
+			return SC_ERROR_INTERNAL;
 	}
 		
 	/* select the application DF */
