@@ -32,28 +32,28 @@
 #define PREALLOCATE	5
 
 /* function declarations */
-static int openct_reader_init(struct sc_context *ctx, void **priv_data);
-static int openct_add_reader(struct sc_context *ctx, unsigned int num, ct_info_t *info);
-static int openct_reader_finish(struct sc_context *ctx, void *priv_data);
-static int openct_reader_release(struct sc_reader *reader);
-static int openct_reader_detect_card_presence(struct sc_reader *reader,
-			struct sc_slot_info *slot);
-static int openct_reader_connect(struct sc_reader *reader,
-			struct sc_slot_info *slot);
-static int openct_reader_disconnect(struct sc_reader *reader,
-			struct sc_slot_info *slot, int action);
-static int openct_reader_transmit(struct sc_reader *reader,
-			struct sc_slot_info *slot,
+static int openct_reader_init(sc_context_t *ctx, void **priv_data);
+static int openct_add_reader(sc_context_t *ctx, unsigned int num, ct_info_t *info);
+static int openct_reader_finish(sc_context_t *ctx, void *priv_data);
+static int openct_reader_release(sc_reader_t *reader);
+static int openct_reader_detect_card_presence(sc_reader_t *reader,
+			sc_slot_info_t *slot);
+static int openct_reader_connect(sc_reader_t *reader,
+			sc_slot_info_t *slot);
+static int openct_reader_disconnect(sc_reader_t *reader,
+			sc_slot_info_t *slot, int action);
+static int openct_reader_transmit(sc_reader_t *reader,
+			sc_slot_info_t *slot,
 			const u8 *sendbuf, size_t sendsize,
 			u8 *recvbuf, size_t *recvsize, unsigned long control);
-static int openct_reader_perform_verify(struct sc_reader *reader,
-			struct sc_slot_info *slot,
+static int openct_reader_perform_verify(sc_reader_t *reader,
+			sc_slot_info_t *slot,
 			struct sc_pin_cmd_data *info);
-static int openct_reader_lock(struct sc_reader *reader,
-			struct sc_slot_info *slot);
-static int openct_reader_unlock(struct sc_reader *reader,
-			struct sc_slot_info *slot);
-static int		openct_error(struct sc_reader *, int);
+static int openct_reader_lock(sc_reader_t *reader,
+			sc_slot_info_t *slot);
+static int openct_reader_unlock(sc_reader_t *reader,
+			sc_slot_info_t *slot);
+static int		openct_error(sc_reader_t *, int);
 
 static struct sc_reader_operations openct_ops;
 
@@ -82,12 +82,21 @@ struct slot_data {
  * is loaded
  */
 static int
-openct_reader_init(struct sc_context *ctx, void **priv_data)
+openct_reader_init(sc_context_t *ctx, void **priv_data)
 {
-	unsigned int	i;
+	unsigned int	i,max;
+	scconf_block *conf_block;
 
 	SC_FUNC_CALLED(ctx, 1);
-	for (i = 0; i < OPENCT_MAX_READERS; i++) {
+
+	max=OPENCT_MAX_READERS; 
+
+        conf_block = _get_conf_block(ctx, "reader_driver", "openct", 1);
+	if (conf_block) {
+		max = scconf_get_int(conf_block, "readers", OPENCT_MAX_READERS);
+	}
+
+	for (i = 0; i < max; i++) {
 		ct_info_t	info;
 
 		if (ct_reader_info(i, &info) >= 0) {
@@ -101,7 +110,7 @@ openct_reader_init(struct sc_context *ctx, void **priv_data)
 }
 
 static int
-openct_add_reader(struct sc_context *ctx, unsigned int num, ct_info_t *info)
+openct_add_reader(sc_context_t *ctx, unsigned int num, ct_info_t *info)
 {
 	sc_reader_t	*reader;
 	struct driver_data *data;
@@ -152,7 +161,7 @@ openct_add_reader(struct sc_context *ctx, unsigned int num, ct_info_t *info)
  * deallocate the private data and any resources.
  */
 int
-openct_reader_finish(struct sc_context *ctx, void *priv_data)
+openct_reader_finish(sc_context_t *ctx, void *priv_data)
 {
 	SC_FUNC_CALLED(ctx, 1);
 	return SC_NO_ERROR;
@@ -164,18 +173,24 @@ openct_reader_finish(struct sc_context *ctx, void *priv_data)
  * freed by OpenSC.
  */
 int
-openct_reader_release(struct sc_reader *reader)
+openct_reader_release(sc_reader_t *reader)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
+	int i;
 
 	SC_FUNC_CALLED(reader->ctx, 1);
 	if (data) {
 		if (data->h)
 			ct_reader_disconnect(data->h);
-		memset(data, 0, sizeof(*data));
+		sc_mem_clear(data, sizeof(*data));
 		reader->drv_data = NULL;
 		free(data);
 	}
+
+	for (i = 0; i < SC_MAX_SLOTS; i++) {
+		if(reader->slot[i].drv_data)
+			free(reader->slot[i].drv_data);
+	}
 	
 	return SC_NO_ERROR;
 }
@@ -184,8 +199,8 @@ openct_reader_release(struct sc_reader *reader)
  * Check whether a card was added/removed
  */
 int
-openct_reader_detect_card_presence(struct sc_reader *reader,
-			struct sc_slot_info *slot)
+openct_reader_detect_card_presence(sc_reader_t *reader,
+			sc_slot_info_t *slot)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 	int rc, status;
@@ -208,8 +223,8 @@ openct_reader_detect_card_presence(struct sc_reader *reader,
 }
 
 static int
-openct_reader_connect(struct sc_reader *reader,
-			struct sc_slot_info *slot)
+openct_reader_connect(sc_reader_t *reader,
+			sc_slot_info_t *slot)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 	int rc;
@@ -243,8 +258,8 @@ openct_reader_connect(struct sc_reader *reader,
 }
 
 static int
-openct_reader_reconnect(struct sc_reader *reader,
-			struct sc_slot_info *slot)
+openct_reader_reconnect(sc_reader_t *reader,
+			sc_slot_info_t *slot)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 	int	rc;
@@ -258,8 +273,8 @@ openct_reader_reconnect(struct sc_reader *reader,
 }
 
 int
-openct_reader_disconnect(struct sc_reader *reader,
-			struct sc_slot_info *slot, int action)
+openct_reader_disconnect(sc_reader_t *reader,
+			sc_slot_info_t *slot, int action)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 
@@ -271,8 +286,8 @@ openct_reader_disconnect(struct sc_reader *reader,
 }
 
 int
-openct_reader_transmit(struct sc_reader *reader,
-		struct sc_slot_info *slot,
+openct_reader_transmit(sc_reader_t *reader,
+		sc_slot_info_t *slot,
 		const u8 *sendbuf, size_t sendsize,
 		u8 *recvbuf, size_t *recvsize, unsigned long control)
 {
@@ -300,8 +315,8 @@ openct_reader_transmit(struct sc_reader *reader,
 }
 
 int
-openct_reader_perform_verify(struct sc_reader *reader,
-		struct sc_slot_info *slot,
+openct_reader_perform_verify(sc_reader_t *reader,
+		sc_slot_info_t *slot,
 		struct sc_pin_cmd_data *info)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
@@ -363,8 +378,8 @@ openct_reader_perform_verify(struct sc_reader *reader,
 
 
 int
-openct_reader_lock(struct sc_reader *reader,
-			struct sc_slot_info *slot)
+openct_reader_lock(sc_reader_t *reader,
+			sc_slot_info_t *slot)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 	struct slot_data *slot_data = (struct slot_data *) slot->drv_data;
@@ -390,8 +405,8 @@ openct_reader_lock(struct sc_reader *reader,
 }
 
 int
-openct_reader_unlock(struct sc_reader *reader,
-			struct sc_slot_info *slot)
+openct_reader_unlock(sc_reader_t *reader,
+			sc_slot_info_t *slot)
 {
 	struct driver_data *data = (struct driver_data *) reader->drv_data;
 	struct slot_data *slot_data = (struct slot_data *) slot->drv_data;
@@ -416,7 +431,7 @@ openct_reader_unlock(struct sc_reader *reader,
  * Handle an error code returned by OpenCT
  */
 int
-openct_error(struct sc_reader *reader, int code)
+openct_error(sc_reader_t *reader, int code)
 {
 	if (code >= 0)
 		return code;
