@@ -7,7 +7,6 @@
 
 #include <string.h>
 #include "sc-pkcs11.h"
-#include "opensc/scrandom.h"
 
 #ifdef HAVE_OPENSSL
 #include <openssl/evp.h>
@@ -62,8 +61,6 @@ sc_pkcs11_register_openssl_mechanisms(struct sc_pkcs11_card *card)
 	sc_pkcs11_register_mechanism(card, &openssl_ripemd160_mech);
 }
 
-static int rng_seeded = 0; 
-
 
 /*
  * Handle OpenSSL digest functions
@@ -123,55 +120,6 @@ sc_pkcs11_openssl_md_release(sc_pkcs11_operation_t *op)
 	op->priv_data = NULL;
 }
 
-CK_RV
-sc_pkcs11_openssl_add_seed_rand(struct sc_pkcs11_session *session,
-				CK_BYTE_PTR pSeed, CK_ULONG ulSeedLen)
-{
-	if (!(session->slot->card->card->caps & SC_CARD_CAP_RNG))
-		return CKR_RANDOM_NO_RNG;
-
-	if (pSeed == NULL || ulSeedLen == 0)
-		return CKR_OK;
-
-	RAND_seed(pSeed, ulSeedLen);
-
-	return CKR_OK;
-}
-
-CK_RV
-sc_pkcs11_openssl_add_gen_rand(struct sc_pkcs11_session *session,
-				CK_BYTE_PTR RandomData, CK_ULONG ulRandomLen)
-{
-	unsigned char seed[20];
-	int r;
-
-	if (!(session->slot->card->card->caps & SC_CARD_CAP_RNG))
-		return CKR_RANDOM_NO_RNG;
-
-	if (RandomData == NULL || ulRandomLen == 0)
-		return CKR_OK;
-
-	if (scrandom_get_data(seed, 20) == -1) {
-		sc_error(context, "scrandom_get_data() failed\n");
-		return CKR_FUNCTION_FAILED;
-	}
-	RAND_seed(seed, 20);
-
-	if (rng_seeded == 0) {
-		r = sc_get_challenge(session->slot->card->card, seed, 20);
-		if (r != 0) {
-			sc_error(context, "sc_get_challenge() returned %d\n", r);
-			return sc_to_cryptoki_error(r, session->slot->card->reader);
-		}
-		rng_seeded = 1;
-	}
-	RAND_seed(seed, 20);
-
-	r = RAND_bytes(RandomData, ulRandomLen);
-
-	return r == 1 ? CKR_OK : CKR_FUNCTION_FAILED;
-}
-
 static int
 do_convert_bignum(sc_pkcs15_bignum_t *dst, BIGNUM *src)
 {
@@ -179,6 +127,8 @@ do_convert_bignum(sc_pkcs15_bignum_t *dst, BIGNUM *src)
 		return 0;
 	dst->len = BN_num_bytes(src);
 	dst->data = (u8 *) malloc(dst->len);
+	if (dst->data == NULL)
+		return 0;
 	BN_bn2bin(src, dst->data);
 	return 1;
 }
