@@ -25,6 +25,65 @@
 #endif
 #include "pkcs11-display.h"
 
+/* Some Netscape/Mozilla-specific stuff:
+ * http://www.opensource.apple.com/darwinsource/10.3/SecurityNssAsn1-11/nssDER/Source/pkcs11n.h */
+
+/*
+ * Netscape-defined object classes
+ * 
+ */
+#define CKO_NETSCAPE 0xCE534350
+
+#define CKO_NETSCAPE_CRL                (CKO_NETSCAPE + 1)
+#define CKO_NETSCAPE_SMIME              (CKO_NETSCAPE + 2)
+#define CKO_NETSCAPE_TRUST              (CKO_NETSCAPE + 3)
+#define CKO_NETSCAPE_BUILTIN_ROOT_LIST  (CKO_NETSCAPE + 4)
+
+/*
+ * Netscape-defined object attributes
+ *
+ */
+
+#define CKA_NETSCAPE 0xCE534350
+
+#define CKA_NETSCAPE_URL                (CKA_NETSCAPE +  1)
+#define CKA_NETSCAPE_EMAIL              (CKA_NETSCAPE +  2)
+#define CKA_NETSCAPE_SMIME_INFO         (CKA_NETSCAPE +  3)
+#define CKA_NETSCAPE_SMIME_TIMESTAMP    (CKA_NETSCAPE +  4)
+#define CKA_NETSCAPE_PKCS8_SALT         (CKA_NETSCAPE +  5)
+#define CKA_NETSCAPE_PASSWORD_CHECK     (CKA_NETSCAPE +  6)
+#define CKA_NETSCAPE_EXPIRES            (CKA_NETSCAPE +  7)
+#define CKA_NETSCAPE_KRL                (CKA_NETSCAPE +  8)
+
+#define CKA_NETSCAPE_PQG_COUNTER        (CKA_NETSCAPE +  20)
+#define CKA_NETSCAPE_PQG_SEED           (CKA_NETSCAPE +  21)
+#define CKA_NETSCAPE_PQG_H              (CKA_NETSCAPE +  22)
+#define CKA_NETSCAPE_PQG_SEED_BITS      (CKA_NETSCAPE +  23)
+
+#define CKA_TRUST (CKA_NETSCAPE + 0x2000)
+
+/* "Usage" key information */
+#define CKA_TRUST_DIGITAL_SIGNATURE     (CKA_TRUST +  1)
+#define CKA_TRUST_NON_REPUDIATION       (CKA_TRUST +  2)
+#define CKA_TRUST_KEY_ENCIPHERMENT      (CKA_TRUST +  3)
+#define CKA_TRUST_DATA_ENCIPHERMENT     (CKA_TRUST +  4)
+#define CKA_TRUST_KEY_AGREEMENT         (CKA_TRUST +  5)
+#define CKA_TRUST_KEY_CERT_SIGN         (CKA_TRUST +  6)
+#define CKA_TRUST_CRL_SIGN              (CKA_TRUST +  7)
+
+/* "Purpose" trust information */
+#define CKA_TRUST_SERVER_AUTH           (CKA_TRUST +  8)
+#define CKA_TRUST_CLIENT_AUTH           (CKA_TRUST +  9)
+#define CKA_TRUST_CODE_SIGNING          (CKA_TRUST + 10)
+#define CKA_TRUST_EMAIL_PROTECTION      (CKA_TRUST + 11)
+#define CKA_TRUST_IPSEC_END_SYSTEM      (CKA_TRUST + 12)
+#define CKA_TRUST_IPSEC_TUNNEL          (CKA_TRUST + 13)
+#define CKA_TRUST_IPSEC_USER            (CKA_TRUST + 14)
+#define CKA_TRUST_TIME_STAMPING         (CKA_TRUST + 15)
+#define CKA_CERT_SHA1_HASH	            (CKA_TRUST + 100)
+#define CKA_CERT_MD5_HASH		        (CKA_TRUST + 101)
+
+
 void print_enum(FILE *f, CK_LONG type, CK_VOID_PTR value, CK_ULONG size, CK_VOID_PTR arg)
 {
   enum_spec *spec = (enum_spec*)arg;
@@ -70,7 +129,7 @@ void print_generic(FILE *f, CK_LONG type, CK_VOID_PTR value, CK_ULONG size, CK_V
 }
 
 #ifdef HAVE_OPENSSL
-void print_dn(FILE *f, CK_LONG type, CK_VOID_PTR value, CK_ULONG size, CK_VOID_PTR arg)
+static void print_dn(FILE *f, CK_LONG type, CK_VOID_PTR value, CK_ULONG size, CK_VOID_PTR arg)
 {
   print_generic(f, type, value, size, arg);
   if(size && value) {
@@ -126,6 +185,10 @@ enum_specs ck_cls_s[] = {
   { CKO_SECRET_KEY       , "CKO_SECRET_KEY       " },
   { CKO_HW_FEATURE       , "CKO_HW_FEATURE       " },
   { CKO_DOMAIN_PARAMETERS, "CKO_DOMAIN_PARAMETERS" },
+  { CKO_NETSCAPE_CRL,              "CKO_NETSCAPE_CRL               " },
+  { CKO_NETSCAPE_SMIME ,           "CKO_NETSCAPE_SMIME             " },
+  { CKO_NETSCAPE_TRUST,            "CKO_NETSCAPE_TRUST             " },
+  { CKO_NETSCAPE_BUILTIN_ROOT_LIST, "CKO_NETSCAPE_BUILTIN_ROOT_LIST" },
   { CKO_VENDOR_DEFINED   , "CKO_VENDOR_DEFINED   " }
 };
 
@@ -485,6 +548,12 @@ type_spec ck_attribute_specs[] = {
   { CKA_OWNER             , "CKA_OWNER            ", print_generic, NULL },
   { CKA_ATTR_TYPES        , "CKA_ATTR_TYPES       ", print_generic, NULL },
   { CKA_TRUSTED           , "CKA_TRUSTED          ", print_generic, NULL },
+  { CKA_CERTIFICATE_CATEGORY, "CKA_CERTIFICATE_CATEGORY ", print_generic, NULL },
+  { CKA_JAVA_MIDP_SECURITY_DOMAIN, "CKA_JAVA_MIDP_SECURITY_DOMAIN ", print_generic, NULL },
+  { CKA_URL               , "CKA_URL              ", print_generic, NULL },
+  { CKA_HASH_OF_SUBJECT_PUBLIC_KEY, "CKA_HASH_OF_SUBJECT_PUBLIC_KEY ", print_generic, NULL },
+  { CKA_HASH_OF_ISSUER_PUBLIC_KEY, "CKA_HASH_OF_ISSUER_PUBLIC_KEY ", print_generic, NULL },
+  { CKA_CHECK_VALUE       , "CKA_CHECK_VALUE      ", print_generic, NULL },
   { CKA_KEY_TYPE          , "CKA_KEY_TYPE         ", print_enum,    ck_key_t },
 #ifdef HAVE_OPENSSL 
   { CKA_SUBJECT           , "CKA_SUBJECT          ", print_dn,      NULL },
@@ -526,13 +595,62 @@ type_spec ck_attribute_specs[] = {
   { CKA_ALWAYS_SENSITIVE  , "CKA_ALWAYS_SENSITIVE ", print_boolean, NULL },
   { CKA_KEY_GEN_MECHANISM , "CKA_KEY_GEN_MECHANISM", print_boolean, NULL },
   { CKA_MODIFIABLE        , "CKA_MODIFIABLE       ", print_boolean, NULL },
+  { CKA_ECDSA_PARAMS      , "CKA_ECDSA_PARAMS     ", print_generic, NULL },
   { CKA_EC_PARAMS         , "CKA_EC_PARAMS        ", print_generic, NULL },
   { CKA_EC_POINT          , "CKA_EC_POINT         ", print_generic, NULL },
   { CKA_SECONDARY_AUTH    , "CKA_SECONDARY_AUTH   ", print_generic, NULL },
   { CKA_AUTH_PIN_FLAGS    , "CKA_AUTH_PIN_FLAGS   ", print_generic, NULL },
+  { CKA_ALWAYS_AUTHENTICATE, "CKA_ALWAYS_AUTHENTICATE ", print_generic, NULL },
+  { CKA_WRAP_WITH_TRUSTED , "CKA_WRAP_WITH_TRUSTED ", print_generic, NULL },
+  { CKA_WRAP_TEMPLATE     , "CKA_WRAP_TEMPLATE    ", print_generic, NULL },
+  { CKA_UNWRAP_TEMPLATE   , "CKA_UNWRAP_TEMPLATE  ", print_generic, NULL },
   { CKA_HW_FEATURE_TYPE   , "CKA_HW_FEATURE_TYPE  ", print_generic, NULL },
   { CKA_RESET_ON_INIT     , "CKA_RESET_ON_INIT    ", print_generic, NULL },
-  { CKA_HAS_RESET         , "CKA_HAS_RESET        ", print_generic, NULL }
+  { CKA_HAS_RESET         , "CKA_HAS_RESET        ", print_generic, NULL },
+  { CKA_PIXEL_X           , "CKA_PIXEL_X          ", print_generic, NULL },
+  { CKA_PIXEL_Y           , "CKA_PIXEL_Y          ", print_generic, NULL },
+  { CKA_RESOLUTION        , "CKA_RESOLUTION       ", print_generic, NULL },
+  { CKA_CHAR_ROWS         , "CKA_CHAR_ROWS        ", print_generic, NULL },
+  { CKA_CHAR_COLUMNS      , "CKA_CHAR_COLUMNS     ", print_generic, NULL },
+  { CKA_COLOR             , "CKA_COLOR            ", print_generic, NULL },
+  { CKA_BITS_PER_PIXEL    , "CKA_BITS_PER_PIXEL   ", print_generic, NULL },
+  { CKA_CHAR_SETS         , "CKA_CHAR_SETS        ", print_generic, NULL },
+  { CKA_ENCODING_METHODS  , "CKA_ENCODING_METHODS ", print_generic, NULL },
+  { CKA_MIME_TYPES        , "CKA_MIME_TYPES       ", print_generic, NULL },
+  { CKA_MECHANISM_TYPE    , "CKA_MECHANISM_TYPE   ", print_generic, NULL },
+  { CKA_REQUIRED_CMS_ATTRIBUTES, "CKA_REQUIRED_CMS_ATTRIBUTES ", print_generic, NULL },
+  { CKA_DEFAULT_CMS_ATTRIBUTES, "CKA_DEFAULT_CMS_ATTRIBUTES ", print_generic, NULL },
+  { CKA_SUPPORTED_CMS_ATTRIBUTES, "CKA_SUPPORTED_CMS_ATTRIBUTES ", print_generic, NULL },
+  { CKA_ALLOWED_MECHANISMS, "CKA_ALLOWED_MECHANISMS ", print_generic, NULL },
+  { CKA_NETSCAPE_URL, "CKA_NETSCAPE_URL(Netsc)                         ", print_generic, NULL },
+  { CKA_NETSCAPE_EMAIL, "CKA_NETSCAPE_EMAIL(Netsc)                     ", print_generic, NULL },
+  { CKA_NETSCAPE_SMIME_INFO, "CKA_NETSCAPE_SMIME_INFO(Netsc)           ", print_boolean, NULL },
+  { CKA_NETSCAPE_SMIME_TIMESTAMP, "CKA_NETSCAPE_SMIME_TIMESTAMP(Netsc) ", print_generic, NULL },
+  { CKA_NETSCAPE_PKCS8_SALT, "CKA_NETSCAPE_PKCS8_SALT(Netsc)           ", print_generic, NULL },
+  { CKA_NETSCAPE_PASSWORD_CHECK, "CKA_NETSCAPE_PASSWORD_CHECK(Netsc)   ", print_generic, NULL },
+  { CKA_NETSCAPE_EXPIRES, "CKA_NETSCAPE_EXPIRES(Netsc)                 ", print_generic, NULL },
+  { CKA_NETSCAPE_KRL, "CKA_NETSCAPE_KRL(Netsc)                         ", print_generic, NULL },
+  { CKA_NETSCAPE_PQG_COUNTER, "CKA_NETSCAPE_PQG_COUNTER(Netsc)         ", print_generic, NULL },
+  { CKA_NETSCAPE_PQG_SEED, "CKA_NETSCAPE_PQG_SEED(Netsc)               ", print_generic, NULL },
+  { CKA_NETSCAPE_PQG_H, "CKA_NETSCAPE_PQG_H(Netsc)                     ", print_generic, NULL },
+  { CKA_NETSCAPE_PQG_SEED_BITS, "CKA_NETSCAPE_PQG_SEED_BITS(Netsc)     ", print_generic, NULL },
+  { CKA_TRUST_DIGITAL_SIGNATURE, "CKA_TRUST_DIGITAL_SIGNATURE(Netsc)   ", print_boolean, NULL },
+  { CKA_TRUST_NON_REPUDIATION, "CKA_TRUST_NON_REPUDIATION(Netsc)       ", print_boolean, NULL },
+  { CKA_TRUST_KEY_ENCIPHERMENT, "CKA_TRUST_KEY_ENCIPHERMENT(Netsc)     ", print_boolean, NULL },
+  { CKA_TRUST_DATA_ENCIPHERMENT, "CKA_TRUST_DATA_ENCIPHERMENT(Netsc)   ", print_boolean, NULL },
+  { CKA_TRUST_KEY_AGREEMENT, "CKA_TRUST_KEY_AGREEMENT(Netsc)           ", print_boolean, NULL },
+  { CKA_TRUST_KEY_CERT_SIGN, "CKA_TRUST_KEY_CERT_SIGN(Netsc)           ", print_boolean, NULL },
+  { CKA_TRUST_CRL_SIGN, "CKA_TRUST_CRL_SIGN(Netsc)                     ", print_boolean, NULL },
+  { CKA_TRUST_SERVER_AUTH, "CKA_TRUST_SERVER_AUTH(Netsc)               ", print_boolean, NULL },
+  { CKA_TRUST_CLIENT_AUTH, "CKA_TRUST_CLIENT_AUTH(Netsc)               ", print_boolean, NULL },
+  { CKA_TRUST_CODE_SIGNING, "CKA_TRUST_CODE_SIGNING(Netsc)             ", print_boolean, NULL },
+  { CKA_TRUST_EMAIL_PROTECTION, "CKA_TRUST_EMAIL_PROTECTION(Netsc)     ", print_boolean, NULL },
+  { CKA_TRUST_IPSEC_END_SYSTEM, "CKA_TRUST_IPSEC_END_SYSTEM(Netsc)     ", print_boolean, NULL },
+  { CKA_TRUST_IPSEC_TUNNEL, "CKA_TRUST_IPSEC_TUNNEL(Netsc)             ", print_boolean, NULL },
+  { CKA_TRUST_IPSEC_USER, "CKA_TRUST_IPSEC_USER(Netsc)                 ", print_boolean, NULL },
+  { CKA_TRUST_TIME_STAMPING, "CKA_TRUST_TIME_STAMPING(Netsc)           ", print_boolean, NULL },
+  { CKA_CERT_SHA1_HASH, "CKA_CERT_SHA1_HASH(Netsc)                     ", print_generic, NULL },
+  { CKA_CERT_MD5_HASH, "CKA_CERT_MD5_HASH(Netsc)                       ", print_generic, NULL },
 };
 
 CK_ULONG ck_attribute_num = sizeof(ck_attribute_specs)/sizeof(type_spec);
@@ -549,7 +667,7 @@ const char *lookup_enum_spec(enum_spec *spec, CK_ULONG value)
   return NULL;
 }
 
-const char *lookup_enum(CK_LONG type, CK_ULONG value)
+const char *lookup_enum(CK_ULONG type, CK_ULONG value)
 {
   CK_ULONG i;
   for(i = 0; ck_types[i].type < ( sizeof(ck_types) / sizeof(enum_spec) ) ; i++) {
@@ -590,7 +708,7 @@ void print_slot_list(FILE *f, CK_SLOT_ID_PTR pSlotList, CK_ULONG ulCount)
 void print_slot_info(FILE *f, CK_SLOT_INFO *info)
 {
   int            i;
-  enum_specs ck_flags[3] = {
+  enum_specs ck_flags[] = {
     { CKF_TOKEN_PRESENT    , "CKF_TOKEN_PRESENT                " },
     { CKF_REMOVABLE_DEVICE , "CKF_REMOVABLE_DEVICE             " },
     { CKF_HW_SLOT          , "CKF_HW_SLOT                      " },
@@ -602,7 +720,7 @@ void print_slot_info(FILE *f, CK_SLOT_INFO *info)
   fprintf(f, "      hardwareVersion:         %d.%d\n",    info->hardwareVersion.major, info->hardwareVersion.minor );
   fprintf(f, "      firmwareVersion:         %d.%d\n",    info->firmwareVersion.major, info->firmwareVersion.minor );
   fprintf(f, "      flags:                   %0lx\n",     info->flags );
-  for(i = 0; i < 3; i++) {
+  for(i = 0; i < sizeof (ck_flags) / sizeof (*ck_flags); i++) {
     if(info->flags & ck_flags[i].type) {
       fprintf(f, "        %s\n", ck_flags[i].name);
     }
@@ -611,8 +729,8 @@ void print_slot_info(FILE *f, CK_SLOT_INFO *info)
 
 void print_token_info(FILE *f, CK_TOKEN_INFO *info)
 {
-  int            i;
-  enum_specs ck_flags[18] = {
+  size_t            i;
+  enum_specs ck_flags[] = {
     { CKF_RNG                          , "CKF_RNG                          " },
     { CKF_WRITE_PROTECTED              , "CKF_WRITE_PROTECTED              " },
     { CKF_LOGIN_REQUIRED               , "CKF_LOGIN_REQUIRED               " },
@@ -651,7 +769,7 @@ void print_token_info(FILE *f, CK_TOKEN_INFO *info)
   fprintf(f, "      firmwareVersion:         %d.%d\n",     info->firmwareVersion.major, info->firmwareVersion.minor );
   fprintf(f, "      time:                   '%16.16s'\n",  info->utcTime );
   fprintf(f, "      flags:                   %0lx\n",      info->flags );
-  for(i = 0; i < 8; i++) {
+  for(i = 0; i < sizeof (ck_flags) / sizeof (*ck_flags); i++) {
     if(info->flags & ck_flags[i].type) {
       fprintf(f, "        %s\n", ck_flags[i].name);
     }
@@ -723,44 +841,59 @@ void print_mech_info(FILE *f, CK_MECHANISM_TYPE type,
 void print_attribute_list(FILE *f, CK_ATTRIBUTE_PTR pTemplate,
 			  CK_ULONG  ulCount)
 {
-  CK_ULONG j, k;
-  for(j = 0; j < ulCount ; j++) {
-    for(k = 0; k < ck_attribute_num; k++) {
-      if(ck_attribute_specs[k].type == pTemplate[j].type) {
-	fprintf(f, "    %s ", ck_attribute_specs[k].name);
-	if(pTemplate[j].pValue) {
-	  ck_attribute_specs[k].display
-	    (f, pTemplate[j].type, pTemplate[j].pValue,
-	     pTemplate[j].ulValueLen,
-	     ck_attribute_specs[k].arg);
-	} else {
-	  fprintf(f, "has size %ld\n", pTemplate[j].ulValueLen); 
+	CK_ULONG j, k;
+	int found;
+
+	for(j = 0; j < ulCount ; j++) {
+		found = 0;
+		for(k = 0; k < ck_attribute_num; k++) {
+			if(ck_attribute_specs[k].type == pTemplate[j].type) {
+				found = 1;
+				fprintf(f, "    %s ", ck_attribute_specs[k].name);
+				if(pTemplate[j].pValue) {
+					ck_attribute_specs[k].display
+					(f, pTemplate[j].type, pTemplate[j].pValue,
+						pTemplate[j].ulValueLen,
+					ck_attribute_specs[k].arg);
+				} else {
+					fprintf(f, "has size %ld\n", pTemplate[j].ulValueLen); 
+				}
+				k = ck_attribute_num;
+			}
+		}
+		if (!found) {
+			fprintf(f, "    CKA_? (0x%08lx)    ", pTemplate[j].type);
+			fprintf(f, "has size %ld\n", pTemplate[j].ulValueLen); 
+		}
 	}
-	k = ck_attribute_num;
-      }
-    }
-  }
 }
 
 void print_attribute_list_req(FILE *f, CK_ATTRIBUTE_PTR pTemplate,
 			      CK_ULONG  ulCount)
 {
   CK_ULONG j, k;
+  int found;
   for(j = 0; j < ulCount ; j++) {
+    found = 0;
     for(k = 0; k < ck_attribute_num; k++) {
       if(ck_attribute_specs[k].type == pTemplate[j].type) {
+	found = 1;
 	fprintf(f, "    %s ", ck_attribute_specs[k].name);
 	fprintf(f, "requested with %ld buffer\n", pTemplate[j].ulValueLen); 
 	k = ck_attribute_num;
       }
     }
+    if (!found) {
+	fprintf(f, "    CKA_? (0x%08lx)    ", pTemplate[j].type);
+	fprintf(f, "requested with %ld buffer\n", pTemplate[j].ulValueLen); 
+    }
   }
 }
 
 void print_session_info(FILE *f, CK_SESSION_INFO *info)
 {
   int            i;
-  enum_specs ck_flags[2] = {
+  enum_specs ck_flags[] = {
     { CKF_RW_SESSION     , "CKF_RW_SESSION                   " },
     { CKF_SERIAL_SESSION , "CKF_SERIAL_SESSION               " }
   };
@@ -768,7 +901,7 @@ void print_session_info(FILE *f, CK_SESSION_INFO *info)
   fprintf(f, "      slotID:                  %ld\n",       info->slotID );
   fprintf(f, "      state:                  '%32.32s'\n",  lookup_enum(STA_T, info->state));
   fprintf(f, "      flags:                   %0lx\n",     info->flags );
-  for(i = 0; i < 2; i++) {
+  for(i = 0; i < sizeof (ck_flags) / sizeof (*ck_flags); i++) {
     if(info->flags & ck_flags[i].type) {
       fprintf(f, "        %s\n", ck_flags[i].name);
     }
