@@ -22,7 +22,7 @@
 #include <string.h>
 #include "sc-pkcs11.h"
 
-struct sc_context *context = NULL;
+sc_context_t *context = NULL;
 struct sc_pkcs11_pool session_pool;
 struct sc_pkcs11_slot virtual_slots[SC_PKCS11_MAX_VIRTUAL_SLOTS];
 struct sc_pkcs11_card card_table[SC_PKCS11_MAX_READERS];
@@ -36,7 +36,7 @@ CK_RV C_Initialize(CK_VOID_PTR pReserved)
 
 	if (context != NULL) {
 		sc_error(context, "C_Initialize(): Cryptoki already initialized\n");
-                return CKR_CRYPTOKI_ALREADY_INITIALIZED;
+		return CKR_CRYPTOKI_ALREADY_INITIALIZED;
 	}
 	rc = sc_establish_context(&context, "opensc-pkcs11");
 	if (rc != 0) {
@@ -48,18 +48,23 @@ CK_RV C_Initialize(CK_VOID_PTR pReserved)
 	load_pkcs11_parameters(&sc_pkcs11_conf, context);
 
 	first_free_slot = 0;
-        pool_initialize(&session_pool, POOL_TYPE_SESSION);
+	pool_initialize(&session_pool, POOL_TYPE_SESSION);
 	for (i=0; i<SC_PKCS11_MAX_VIRTUAL_SLOTS; i++)
-                slot_initialize(i, &virtual_slots[i]);
+		slot_initialize(i, &virtual_slots[i]);
 	for (i=0; i<SC_PKCS11_MAX_READERS; i++)
-                card_initialize(i);
+		card_initialize(i);
 
 	/* Detect any card, but do not flag "insert" events */
 	__card_detect_all(0);
 
 	rv = sc_pkcs11_init_lock((CK_C_INITIALIZE_ARGS_PTR) pReserved);
+	if (rv != CKR_OK)   {
+		sc_release_context(context);
+		context = NULL;
+	}
 
-out:	if (context != NULL)
+out:	
+	if (context != NULL)
 		sc_debug(context, "C_Initialize: result = %d\n", rv);
 	return rv;
 }
@@ -79,8 +84,8 @@ CK_RV C_Finalize(CK_VOID_PTR pReserved)
 	}
 
 	sc_debug(context, "Shutting down Cryptoki\n");
-	for (i=0; i<context->reader_count; i++)
-                card_removed(i);
+	for (i=0; i < (int)sc_ctx_get_reader_count(context); i++)
+		card_removed(i);
 
 	sc_release_context(context);
 	context = NULL;
@@ -88,7 +93,7 @@ CK_RV C_Finalize(CK_VOID_PTR pReserved)
 out:	/* Release and destroy the mutex */
 	sc_pkcs11_free_lock();
 
-        return rv;
+	return rv;
 }
 
 CK_RV C_GetInfo(CK_INFO_PTR pInfo)
@@ -113,13 +118,13 @@ CK_RV C_GetInfo(CK_INFO_PTR pInfo)
 		  "OpenSC Project (www.opensc.org)",
 		  sizeof(pInfo->manufacturerID));
 	strcpy_bp(pInfo->libraryDescription,
-		  "SmartCard PKCS#11 API",
+		  "smart card PKCS#11 API",
 		  sizeof(pInfo->libraryDescription));
 	pInfo->libraryVersion.major = 1;
 	pInfo->libraryVersion.minor = 0;
 
 out:	sc_pkcs11_unlock();
-        return rv;
+	return rv;
 }	
 
 CK_RV C_GetFunctionList(CK_FUNCTION_LIST_PTR_PTR ppFunctionList)
@@ -136,7 +141,8 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 		    CK_ULONG_PTR   pulCount)      /* receives the number of slots */
 {
 	CK_SLOT_ID found[SC_PKCS11_MAX_VIRTUAL_SLOTS];
-	int numMatches, i;
+	int i;
+	CK_ULONG numMatches;
 	sc_pkcs11_slot_t *slot;
 	CK_RV rv;
 
@@ -163,14 +169,14 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	if (pSlotList == NULL_PTR) {
 		sc_debug(context, "was only a size inquiry (%d)\n", numMatches);
 		*pulCount = numMatches;
-                rv = CKR_OK;
+		rv = CKR_OK;
 		goto out;
 	}
 
 	if (*pulCount < numMatches) {
 		sc_debug(context, "buffer was too small (needed %d)\n", numMatches);
 		*pulCount = numMatches;
-                rv = CKR_BUFFER_TOO_SMALL;
+		rv = CKR_BUFFER_TOO_SMALL;
 		goto out;
 	}
 
@@ -181,14 +187,14 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	sc_debug(context, "returned %d slots\n", numMatches);
 
 out:	sc_pkcs11_unlock();
-        return rv;
+	return rv;
 }
 
 CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 {
 	struct sc_pkcs11_slot *slot;
 	sc_timestamp_t now;
-        CK_RV rv;
+	CK_RV rv;
 
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
@@ -207,7 +213,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 		if (now >= card_table[slot->reader].slot_state_expires || now == 0) {
 			/* Update slot status */
 			rv = card_detect(slot->reader);
-			/* Don't ask again within the next second */ 
+			/* Don't ask again within the next second */
 			card_table[slot->reader].slot_state_expires = now + 1000;
 		}
 	}
@@ -224,7 +230,7 @@ out:	sc_pkcs11_unlock();
 CK_RV C_GetTokenInfo(CK_SLOT_ID slotID, CK_TOKEN_INFO_PTR pInfo)
 {
 	struct sc_pkcs11_slot *slot;
-        CK_RV rv;
+	CK_RV rv;
 
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
@@ -250,7 +256,7 @@ CK_RV C_GetMechanismList(CK_SLOT_ID slotID,
                          CK_ULONG_PTR pulCount)
 {
 	struct sc_pkcs11_slot *slot;
-        CK_RV rv;
+	CK_RV rv;
 
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
@@ -269,7 +275,7 @@ CK_RV C_GetMechanismInfo(CK_SLOT_ID slotID,
 			 CK_MECHANISM_INFO_PTR pInfo)
 {
 	struct sc_pkcs11_slot *slot;
-        CK_RV rv;
+	CK_RV rv;
 
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
@@ -295,7 +301,7 @@ CK_RV C_InitToken(CK_SLOT_ID slotID,
 	struct sc_pkcs11_pool_item *item;
 	struct sc_pkcs11_session *session;
 	struct sc_pkcs11_slot *slot;
-        CK_RV rv;
+	CK_RV rv;
 
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
@@ -334,12 +340,24 @@ CK_RV C_WaitForSlotEvent(CK_FLAGS flags,   /* blocking/nonblocking flag */
 			 CK_SLOT_ID_PTR pSlot,  /* location that receives the slot ID */
 			 CK_VOID_PTR pReserved) /* reserved.  Should be NULL_PTR */
 {
-	struct sc_reader *reader, *readers[SC_MAX_SLOTS * SC_MAX_READERS];
+	sc_reader_t *reader, *readers[SC_MAX_SLOTS * SC_MAX_READERS];
 	int slots[SC_MAX_SLOTS * SC_MAX_READERS];
 	int i, j, k, r, found;
 	unsigned int mask, events;
 	CK_RV rv;
 
+	/* Firefox 1.5 calls this function (blocking) from a seperate thread,
+	 * which gives 2 problems:
+	 * - on Windows/Mac: this waiting thread will log to a NULL context
+	 *   after the 'main' thread does a C_Finalize() and sets the ctx to NULL.
+	 * - on Linux, things just hang (at least on Debian 'sid')
+	 * So we just return CKR_FUNCTION_NOT_SUPPORTED on a blocking call,
+	 * in which case Ff just seems to default to polling in the main thread
+	 * as the Mozilla family of browsers did before.
+	 */
+	if (!(flags & CKF_DONT_BLOCK))
+		return CKR_FUNCTION_NOT_SUPPORTED;
+
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
 		return rv;
@@ -355,8 +373,12 @@ CK_RV C_WaitForSlotEvent(CK_FLAGS flags,   /* blocking/nonblocking flag */
 	 || (flags & CKF_DONT_BLOCK))
 		goto out;
 
-	for (i = k = 0; i < context->reader_count; i++) {
-		reader = context->reader[i];
+	for (i = k = 0; i < (int)sc_ctx_get_reader_count(context); i++) {
+		reader = sc_ctx_get_reader(context, i);
+		if (reader == NULL) {
+			rv = CKR_GENERAL_ERROR;
+			goto out;
+		}
 		for (j = 0; j < reader->slot_count; j++, k++) {
 			readers[k] = reader;
 			slots[k] = j;
@@ -498,7 +520,7 @@ sc_pkcs11_free_lock()
 	_lock = NULL;
 
 	/* Now unlock. On SMP machines the synchronization
-	 * primitives should take care of flushing out 
+	 * primitives should take care of flushing out
 	 * all changed data to RAM */
 	__sc_pkcs11_unlock(tempLock);
 
@@ -548,7 +570,7 @@ CK_FUNCTION_LIST pkcs11_function_list = {
 	C_Decrypt,
 	C_DecryptUpdate,
 	C_DecryptFinal,
-        C_DigestInit,
+	C_DigestInit,
 	C_Digest,
 	C_DigestUpdate,
 	C_DigestKey,
@@ -577,6 +599,6 @@ CK_FUNCTION_LIST pkcs11_function_list = {
 	C_SeedRandom,
 	C_GenerateRandom,
 	C_GetFunctionStatus,
-        C_CancelFunction,
+	C_CancelFunction,
 	C_WaitForSlotEvent
 };
