@@ -1,5 +1,5 @@
 /*
- * slot.c: Smartcard and slot related management functions
+ * slot.c: smart card and slot related management functions
  *
  * Copyright (C) 2002  Timo Ters <timo.teras@iki.fi>
  *
@@ -22,11 +22,11 @@
 #include "sc-pkcs11.h"
 
 static struct sc_pkcs11_framework_ops *frameworks[] = {
-        &framework_pkcs15,
+	&framework_pkcs15,
 #ifdef USE_PKCS15_INIT
 	/* This should be the last framework, because it
 	 * will assume the card is blank and try to initialize it */
-        &framework_pkcs15init,
+	&framework_pkcs15init,
 #endif
 	NULL
 };
@@ -41,7 +41,7 @@ static void init_slot_info(CK_SLOT_INFO_PTR pInfo)
 	pInfo->hardwareVersion.major = 0;
 	pInfo->hardwareVersion.minor = 0;
 	pInfo->firmwareVersion.major = 0;
-        pInfo->firmwareVersion.minor = 0;
+	pInfo->firmwareVersion.minor = 0;
 }
 
 CK_RV card_initialize(int reader)
@@ -70,23 +70,25 @@ CK_RV card_initialize(int reader)
 	card->num_slots = 0;
 
 	first_free_slot += card->max_slots;
-        return CKR_OK;
+	return CKR_OK;
 }
 
 CK_RV card_detect(int reader)
 {
 	struct sc_pkcs11_card *card = &card_table[reader];
-        int rc, rv, i, retry = 1;
+	int rc, rv, i, retry = 1;
 
-        rv = CKR_OK;
+	rv = CKR_OK;
 
-	sc_debug(context, "%d: Detecting SmartCard\n", reader);
+	sc_debug(context, "%d: Detecting smart card\n", reader);
 	for (i = card->max_slots; i--; ) {
 		struct sc_pkcs11_slot *slot;
+		sc_reader_t *rdr = sc_ctx_get_reader(context, (unsigned int)reader);
 
+		if (rdr == NULL)
+			return CKR_GENERAL_ERROR;
 		slot = virtual_slots + card->first_slot + i;
-		strcpy_bp(slot->slot_info.slotDescription,
-				context->reader[reader]->name, 64);
+		strcpy_bp(slot->slot_info.slotDescription, rdr->name, 64);
 		slot->reader = reader;
 	}
 
@@ -118,8 +120,8 @@ again:	rc = sc_detect_card_presence(context->reader[reader], 0);
 
 	/* Detect the card if it's not known already */
 	if (card->card == NULL) {
-		sc_debug(context, "%d: Connecting to SmartCard\n", reader);
-		rc = sc_connect_card(context->reader[reader], 0, &card->card);
+		sc_debug(context, "%d: Connecting to smart card\n", reader);
+		rc = sc_connect_card(sc_ctx_get_reader(context, reader), 0, &card->card);
 		if (rc != SC_SUCCESS)
 			return sc_to_cryptoki_error(rc, reader);
 	}
@@ -143,7 +145,7 @@ again:	rc = sc_detect_card_presence(context->reader[reader], 0);
 		sc_debug(context, "%d: Detected framework %d. Creating tokens.\n", reader, i);
 		rv = frameworks[i]->create_tokens(card);
 		if (rv != CKR_OK)
-                        return rv;
+			return rv;
 
 		card->framework = frameworks[i];
 	}
@@ -158,7 +160,7 @@ CK_RV __card_detect_all(int report_events)
 
 	if (context == NULL_PTR)
 		return CKR_CRYPTOKI_NOT_INITIALIZED;
-	for (i = 0; i < context->reader_count; i++)
+	for (i = 0; i < (int)sc_ctx_get_reader_count(context); i++)
 		card_detect(i);
 	if (!report_events) {
 		CK_SLOT_ID id;
@@ -178,14 +180,14 @@ CK_RV card_detect_all(void)
 CK_RV card_removed(int reader)
 {
 	int i;
-        struct sc_pkcs11_card *card;
+	struct sc_pkcs11_card *card;
 
-	sc_debug(context, "%d: SmartCard removed\n", reader);
+	sc_debug(context, "%d: smart card removed\n", reader);
 
 	for (i=0; i<SC_PKCS11_MAX_VIRTUAL_SLOTS; i++) {
 		if (virtual_slots[i].card &&
 		    virtual_slots[i].card->reader == reader)
-                        slot_token_removed(i);
+				slot_token_removed(i);
 	}
 
 	/* beware - do not clean the entire sc_pkcs11_card struct;
@@ -200,9 +202,9 @@ CK_RV card_removed(int reader)
 
 	if (card->card)
 		sc_disconnect_card(card->card, 0);
-        card->card = NULL;
+	card->card = NULL;
 
-        return CKR_OK;
+	return CKR_OK;
 }
 
 CK_RV slot_initialize(int id, struct sc_pkcs11_slot *slot)
@@ -210,10 +212,10 @@ CK_RV slot_initialize(int id, struct sc_pkcs11_slot *slot)
 	memset(slot, 0, sizeof(*slot));
 	slot->id = id;
 	slot->login_user = -1;
-        init_slot_info(&slot->slot_info);
+	init_slot_info(&slot->slot_info);
 	pool_initialize(&slot->object_pool, POOL_TYPE_OBJECT);
 
-        return CKR_OK;
+	return CKR_OK;
 }
 
 CK_RV slot_allocate(struct sc_pkcs11_slot **slot, struct sc_pkcs11_card *card)
@@ -228,28 +230,26 @@ CK_RV slot_allocate(struct sc_pkcs11_slot **slot, struct sc_pkcs11_card *card)
 	for (i = first; i < last; i++) {
 		if (!virtual_slots[i].card) {
 			sc_debug(context, "Allocated slot %d\n", i);
-
-                        virtual_slots[i].card = card;
-                        virtual_slots[i].events = SC_EVENT_CARD_INSERTED;
+			virtual_slots[i].card = card;
+			virtual_slots[i].events = SC_EVENT_CARD_INSERTED;
 			*slot = &virtual_slots[i];
 			card->num_slots++;
 			return CKR_OK;
 		}
 	}
-        return CKR_FUNCTION_FAILED;
-
+	return CKR_FUNCTION_FAILED;
 }
 
 CK_RV slot_get_slot(int id, struct sc_pkcs11_slot **slot)
 {
 	if (context == NULL)
-                return CKR_CRYPTOKI_NOT_INITIALIZED;
+		return CKR_CRYPTOKI_NOT_INITIALIZED;
 
 	if (id < 0 || id >= SC_PKCS11_MAX_VIRTUAL_SLOTS)
 		return CKR_SLOT_ID_INVALID;
-
-        *slot = &virtual_slots[id];
-        return CKR_OK;
+		
+	*slot = &virtual_slots[id];
+	return CKR_OK;
 }
 
 CK_RV slot_get_token(int id, struct sc_pkcs11_slot **slot)
@@ -266,15 +266,14 @@ CK_RV slot_get_token(int id, struct sc_pkcs11_slot **slot)
 		if (rv != CKR_OK)
 			return CKR_TOKEN_NOT_PRESENT;
 	}
-
-        return CKR_OK;
+	return CKR_OK;
 }
 
 CK_RV slot_token_removed(int id)
 {
 	int rv, token_was_present;
-        struct sc_pkcs11_slot *slot;
-        struct sc_pkcs11_object *object;
+	struct sc_pkcs11_slot *slot;
+	struct sc_pkcs11_object *object;
 	CK_SLOT_INFO saved_slot_info;
 	int reader;
 
@@ -284,12 +283,12 @@ CK_RV slot_token_removed(int id)
 
 	token_was_present = (slot->slot_info.flags & CKF_TOKEN_PRESENT);
 
-        /* Terminate active sessions */
-        sc_pkcs11_close_all_sessions(id);
+	/* Terminate active sessions */
+	sc_pkcs11_close_all_sessions(id);
 
 	/* Object pool */
 	while (pool_find_and_delete(&slot->object_pool, 0, (void**) &object) == CKR_OK) {
-                if (object->ops->release)
+		if (object->ops->release)
 			object->ops->release(object);
 	}
 
@@ -316,8 +315,7 @@ CK_RV slot_token_removed(int id)
 	if (token_was_present)
 		slot->events = SC_EVENT_CARD_REMOVED;
 
-        return CKR_OK;
-
+	return CKR_OK;
 }
 
 CK_RV slot_find_changed(CK_SLOT_ID_PTR idp, int mask)
